# - Find Git
# Find if git is installed on the system. It will give back the following variable
#
# GIT_EXECUTABLE      - Full path to the git executable
# GIT_VERSION_STRING  - Human readable version string
# GIT_VERSION_MAJOR   - Major version of git
# GIT_VERSION_MINOR   - Minor version of git
# GIT_VERSION_PATCH   - Patch version of git
# GIT_FOUND           - True if Git has been found
#
# Created on : 9 jun. 2010
# author : Alexandre Tuleu
# Copyright 2010 Alexandre Tuleu
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

if(GIT_EXECUTABLE)
  set(GIT_FIND_QUIETLY TRUE)
endif(GIT_EXECUTABLE)

find_program(GIT_EXECUTABLE git)

set(MESSAGE_TYPE )# if message type is void we don't send message

if(GIT_FIND_REQUIRED)#if required message are error
  set(MESSAGE_TYPE FATAL_ERROR)
else(GIT_FIND_REQUIRED)
  if(NOT GIT_FIND_QUIETLY) # if not required and not quiet message are warning
    set(MESSAGE_TYPE STATUS)
  endif(NOT GIT_FIND_QUIETLY)
endif(GIT_FIND_REQUIRED)

#get the version from git itself
if(GIT_EXECUTABLE)
  execute_process(COMMAND ${GIT_EXECUTABLE} "--version" OUTPUT_VARIABLE 
                          GIT_VERSION_TMP OUTPUT_STRIP_TRAILING_WHITESPACE)
  string(REGEX REPLACE ".* ([0-9]+)\\.([0-9]+)\\.([0-9]+).*" "\\1" 
                       GIT_VERSION_MAJOR ${GIT_VERSION_TMP})
  string(REGEX REPLACE ".* ([0-9]+)\\.([0-9]+)\\.([0-9]+).*" "\\2" 
                       GIT_VERSION_MINOR ${GIT_VERSION_TMP})
  string(REGEX REPLACE ".* ([0-9]+)\\.([0-9]+)\\.([0-9]+).*" "\\3" 
                       GIT_VERSION_PATCH ${GIT_VERSION_TMP})
  set(GIT_VERSION_STRING "${GIT_VERSION_MAJOR}.${GIT_VERSION_MINOR}.${GIT_VERSION_PATCH}")
endif(GIT_EXECUTABLE)


#strandard arguments
include(FindPackageHandleStandardArgs)
find_package_handle_standard_args(GIT DEFAULT_MSG GIT_EXECUTABLE 
                                  GIT_VERSION_STRING)
                                  
#Test for the right version of GIT according to the standard FindXXX.cmake arguments
set(VERSION_MATCH TRUE)
if(GIT_FIND_VERSION_EXACT)#user want exact version
  if(NOT GIT_VERSION_STRING VERSION_EQUAL GIT_FIND_VERSION)#it's diffrent
    set(VERSION_MATCH FALSE)
  endif(NOT GIT_VERSION_STRING VERSION_EQUAL GIT_FIND_VERSION)
elseif(GIT_FIND_VERSION_EXACT) #user want just at least the indicated version
  if(GIT_VERSION_STRING VERSION_LESS GIT_FIND_VERSION) # too low
    set(VERSION_MATCH FALSE)
  endif(GIT_VERSION_STRINg VERSION_LESS GIT_FIND_VERSION)
endif(GIT_FIND_VERSION_EXACT)

#user want a specific version and we didnot found one
if(GIT_FIND_VERSION AND NOT VERSION_MATCH)
  if(MESSAGE_TYPE)# need to send a message
    message(${MESSAGE_TYPE} "Could NOT find Git (found version ${GIT_VERSION_STRING}, but desired version is ${GIT_FIND_VERSION})")
  endif(MESSAGE_TYPE)
  set(GIT_FOUND NOTFOUND) # we don't have found what we wanted
endif(GIT_FIND_VERSION AND NOT VERSION_MATCH)

#user don't have to see everything
mark_as_advanced(GIT_EXECUTABLE GIT_VERSION_MAJOR GIT_VERSION_MINOR 
                 GIT_VERSION_PATCH GIT_VERSION_STRING)
                 
                 
# Macro for releasing sources package using git. When you use git, you specify 
# a lot of (all) files that are generated by cmake or your compiler, and that are
# not intended to be tracked. If you use CPack for releasing sources tarball and
# git as SCM, you must list these files twice, one for git, and one for CPack.
# With this macro, we only use the git '.gitignore' system, with the 
# 'git-archive' command to generate the sources tarball. The source-tarball 
# target/solution is added to the build system, that will generate the tarball.              
#
# usage USE_GIT_FOR_SOURCE_TARBALL_TARGET([DESIRED_VERSION])
# input : 
#  DESIRED_VERSION : if specified add "-${DESIRED_VERSION}" suffix to the files
# output : 
#  add the 'source-tarball' make target/solution to the build system 
             
function(USE_GIT_FOR_SOURCE_TARBALL_TARGET DESIRED_VERSION)
  if(GIT_FOUND AND NOT GIT_VERSION_STRING VERSION_LESS 1.4.4) # git command 'archive' introduced in 1.4.4
    set(tarball_basename ${PROJECT_NAME})
    if(DESIRED_VERSION) #Want to add the version suffix
      set(tarball_basename "${tarball_basename}-${DESIRED_VERSION}")
    endif(DESIRED_VERSION)
    set(GIT_ARCHIVE_COMMAND_ARGS "archive" "--prefix=${tarball_basename}/"
                                 "HEAD")
    find_program(GZIP_FOUND gzip) # its always better to compress, but it may not be possible
    if(GZIP_FOUND) # we can compress, so we produce tar.gz
      set(tarball_name "${tarball_basename}.tar.gz" )
      set(GIT_ARCHIVE_COMMAND_ARGS ${GIT_ARCHIVE_COMMAND_ARGS}
                                   "|" "gzip" ">${tarball_name}")
    else(GZIP_FOUND)# we cannot compress so we create a a simple .tar
      set(tarball_name "${tarball_basename}.tar" )   
      set(GIT_ARCHIVE_COMMAND_ARGS "--format=tar" "--output=${tarball_name}"
                                   ${GIT_ARCHIVE_COMMAND_ARGS})
    endif(GZIP_FOUND)
    #create the custom target
    message(STATUS "Added source-tarball target/solution that will create ${tarball_name} sources release.")
    add_custom_target(source-tarball 
                      COMMAND ${GIT_EXECUTABLE} ${GIT_ARCHIVE_COMMAND_ARGS}
                      COMMENT "Creating the file ${tarball_name}")
  endif(GIT_FOUND AND NOT GIT_VERSION_STRING VERSION_LESS 1.4.4)
endfunction(USE_GIT_FOR_SOURCE_TARBALL_TARGET)
