/*
 * Copyright (c) 2008-2010 Mikaël Mayer, Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file         emitter_receiver_physics.hpp
 * \date         15/11/2008
 * \brief         Implements an emitter/receiver for Webots physic plug-in
 * \author       Mikaël Mayer, Alexandre Tuleu
 */

#ifndef _EMITTER_RECEIVER_PHYSICS_HPP_
#define _EMITTER_RECEIVER_PHYSICS_HPP_
#define  _ODE_ODECPP_H_
#include <ode/ode.h>
#include <plugins/physics.h>
#include "MessageException.h"
#include "message_private.hpp"
#include <map>
/**
 * \class  EmitterReceiverPhysics
 * \author Mikaël Mayer
 * \date   2008/11/15
 * Implements the EmitterReceiver interface
 * with the emitter and receivers of a physic plug-in.
 *
 * However, like webots is not very nice, physics plug-ins doesn't have the same
 * interface than Robot and Supervisor : they receive raw data rather than
 * individual packet. They also send raw data. A workaround has been made for
 * sending message, but this is not the case for the sending. Therefore you should never
 * send more than one message per simulation step with this class (ie. inside a
 * physics plugins).
 *
 * \attention before to include the traditional message.hpp headers, you need to
 * define the preprocessor variable  FOR_PHYSICS_PLUGIN :
\code
#define FOR_PHYSICS_PLUGIN
#include <Message/message.hpp>
\endcode
 * \ingroup Message
 * \todo make a workarround for the webots limitation, not to be able to send
 *       one message.
 */
class EmitterReceiverPhysics:public EmitterReceiver {

private:
  int channel;
  bool data_available;
  bool initialized_;
  std::map<MessageType,unsigned int> sizes_;

  std::list<std::pair<int,unsigned char*> > datas_;



  void initialize(){
    for(std::list<MessageType>::const_iterator it=InternalMessage::registeredTypes_->begin();
        it!=InternalMessage::registeredTypes_->end();
        ++it){
      unsigned int id=static_cast<unsigned int>(*it);
      MESSAGE_CREATOR creat =InternalMessage::registered_types[id];
      InternalMessage *m=(*creat)();
      int size=0;
      unsigned char *data= NULL;
      m->serialize_complete(&data, &size);
      sizes_.insert(std::make_pair(*it,size));
    }
    initialized_=true;
  }

public:

  /// The constructor
  EmitterReceiverPhysics(int channel_ex=0)
    : channel(channel_ex)
    , data_available(false)
    , initialized_(false){

  }

  void check_packet() {
    if(!initialized_)
      initialize();
    if(data_available) return;
    int size;
    void *data = dWebotsReceive(&size);
    if(data==NULL) return;

    int data_parsed=0;
    //NOW WE HAVE TO SEPARATE EACH MESSAGE

    const unsigned char* ptr= (const unsigned char*) data;

    while(data_parsed<size){
      int sizePacket=sizes_[static_cast<MessageType>(ptr[0])];
      unsigned char* dd= new unsigned char[sizePacket];
      memcpy(dd,ptr,sizePacket);
      data_parsed+=sizePacket;
      ptr+=sizePacket;
      datas_.push_back(std::make_pair(sizePacket,dd));
    }
    //message separated
    data_available = true;
  }

  void pop_head(){
    if(datas_.empty())
      return;
    delete[] datas_.front().second;
    datas_.pop_front();
    data_available=!datas_.empty();
  }

public:
  /// The default construct procedure
  inline static EmitterReceiver* create(int channel=0) {
    return new EmitterReceiverPhysics(channel);
  }

  /// Emitter methods


  inline virtual int sendPacket(const void *data, int size) {
    #ifdef LINUX
      dWebotsSend(channel, data, size);
    #else
      char* tmp=new char[size];
      memcpy(tmp,data,size);
      dWebotsSend(channel,tmp,size);//Webots bug for physics plugin
      //need a void* but only used as a const. only const void* defined in
      //cyberbotic doc.
      delete[] tmp;
    #endif
    return 0;
  }

  /// Receiver methods
  inline virtual void enable(unsigned short ) {
  }
  inline virtual int getQueueLength() {
    check_packet();
    return datas_.size();
  }
  inline virtual void nextPacket() {
    //check_packet();
    pop_head();

  }
  inline virtual int getDataSize() {
    check_packet();
    return datas_.front().first;
  }
  inline virtual const void *getData() {
    check_packet();
    return datas_.front().second;
  }
};

#endif /*_EMITTER_RECEIVER_PHYSICS_HPP_*/
