/*
 * Copyright (c) 2008-2010 Mikaël Mayer
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file         message.hpp
 * \date         17/11/2008
 * \brief        Default common messages
 * \author       Mikaël Mayer
 * \brief Common message.hpp header.
 * You should include this only one header for an external use of the library.
 */


/*
 * \mainpage
 * libMessage is a librairy for webots, that set up an Object Oriented Mesaging
 * Protocol. Users can use the build-in communication protocol,
 * or define their own message protocol.
 *
 * \section useLibMessage using the libMessage.
 *
 * In order to use the library to exchange message between several
 * controller/supervisor/plug-ins you should provide the following :
 * - each actor on the network should have an webots::Emitter and a webots::Receiver,
 * their channel must be all the same, and the buffer should be long enough to
 * store all the data
 * - each actor of the network should have an int that identifiate him. Different
 * actors could share the same identificator, and respond to different identificator, so
 * you could made group of actor.
 *
 *
 * Then you can receive Message like it is explained in the Message definition class.
 * You could also easily send new Message by the following :
\code
  DummyMessage my_message(MY_ID,DEST_ID);//create a nw message from me to DEST_ID
  my_message.setData(custom_data)//give data to message, depend of the type of message
  my_message.send_me();//send the message over the network.
\endcode
 *
 * \subsection howToInclude Including the library
 * you should only include one header from the library : "Message/message.hpp"
 * In the case of the use inside a physics plugin, you need to define the
 * preprocessor variable FOR_PHYSICS_PLUGIN <i>before</i> including it :
\code
#define FOR_PHYSICS_PLUGIN
#include <Message/message.hpp>
\endcode
 * You should never do this outside a physics plugins (it need the ODE headers,
 * which are provided for physics plugins).
 *
 * \section creating new Message.
 *
 * To do this, you can do the following :
 *  - Create new type of message with a new enum :
\code
#include "Message/message.hpp"

enum CustomMessageType{
  DUMMY_MESSAGE=static_cast<unsigned int>(NB_REGISTERED_MESSAGE), //start to the right position, not t collide with the library message
  DUMMY_MESSAGE2,
  DUMMY_MESSAGE3
};
\endcode
 * - use the IS_A_MESSAGE macro for fast creating some new messages.
 *
 * \subsection IS_A_MESSAGE IS_A_MESSAGE macro
 * A useful macro for defining a new Message.
 * It automatically instanciate :
 *   - the registered boolean
 *   - the create(), clone() and cast() methods
 *   - the size() and type() methods;
 *   - a standard constructor, taking only the ids of the sender and
 *     the receiver.
 *
 *     this macro takes two arguments :
 *     - MessageClassName the name of the class.
 *     - MessageEnumName its asociated MessageType.
 *
 *
 * This macro should be used like this  in the header :
\code
class DummyMessage : public InternalMessage{
IS_A_MESSAGE(DummyMessage, DUMMY_MESSAGE);
protected :
  //define here the data you want to access, here we have a void* pointer
  void * data_;
public :
  ~DummyMessage(){// don't forget RAI principles!�
    delete data_;
  }
  //getter and setters
  void setData(void * data){
    data_=data;
  }
  const void * getData() const{
    return data_;
  }
};
\endcode
 * and don't forget to put in a .cpp file :
\code
  bool DummyMessage::registered=InternalMessage::register_type(
    static_cast<MessageType>(DUMMY_MESSAGE),
    & DummyMessage::create);
\endcode
 *
 *
 * \section limits limitations
 *
 * There is one big limitation with this library : EmitterReceiverPhysics can
 * only send one single Message par simulation step. If you try to send more,
 * it will just crash. Should be corrected in further improvement.
 *
 *
 */

#ifndef DEFAULT_MESSAGES_HPP_
#define DEFAULT_MESSAGES_HPP_

#include "message_private.hpp"


#ifdef FOR_PHYSICS_PLUGIN
  #include "emitter_receiver_physics.hpp"
#else
  #include "emitter_receiver_robot.hpp"
#endif


//#ifndef CUSTOM_COMMUNICATION_PROTOCOL

/// Contains the enum of the messages.
#include "registered_messages.hpp"


/// Default messages:
//#include "messages_positions.hpp"
//#include "messages_lock.hpp"



/// Messages needing some objects.
//#include "messages_referential.hpp"
//#include "messages_CPG.hpp"

///TCP message
//#include "messages_Network.hpp"
//#endif
#endif /* DEFAULT_MESSAGES_HPP_ */
