/*
 * Copyright (c) 2008-2010 Mikaël Mayer
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \date          14/10/2008
 * \file          message_private.hpp
 * \brief Protocol to communicate from different different webots controllers/supervisor/plugins
 *
 * \details Two classes are defined here.
 *
 * First class Message, used for receiving messages and information about them.
 *
 * Second class InternalMessage, an abstract class that has to be inherited
 * to use your own messages.
 *
 * To use them, the robot emitter and receiver used have to be called
 * respectively 'emitter' and 'receiver' without the quotes.
 * To send messages in string human-readable format, use #define MESSAGE_STRING
 * \author:        Mikaël Mayer
 * \note Modifications: Alexandre Tuleu - 21/10/2008 :
 *                    - Correct some variable names in order to respect BIRG coding
 *                      convention.
 */

#ifndef MESSAGE_HPP
#define MESSAGE_HPP

#include <iostream>

#include "serializer.hpp"
#include "registered_messages.hpp"
#include <list>

/**
 * \defgroup Message Message module
 *
 * This is all the documentation about classes that eases the creations of
 * protocols between webots controllers, supervisor and plugins.
 * \ingroup internal
 */

//@{
/**
 * A useful macro for defining new Message.
 *
 * Ideally a new Messages should be defined like this, First add new type :
\code
  #define CUSTOM_COMMUNICATION_PROTOCOL
  #include "message.hpp"

  enum MyMessageType {
    MESSAGE1= static_cast<unsigned int>(NB_REGISTERED_MESSAGE),
    MESSAGE2
  };
\endcode
 * Now we define new Object that correspond to this new type using the macro.
\code
  Message1 : public InternalMessage {
    IS_A_MESSAGE(Message1,MESSAGE1);
  private :
  //write some fields
  double foo_,bar_;
  public :
  //write getters and setters
  double foo() const { return foo_; }
  double bar() const { return bar_; }
  void setFoo(double v){ foo_=v; }
  void setBar(double v){ bar_=v; }

  //(un)serialization functions
  virtual Serializer & serialize(Serializer &s){
    return s<<truc<<bidule;//add your field here
  }

  virtual Serializer & unserialize(Serializer &s){
    return s>>truc>>bidule;//add your field here
  }

  };
\endcode
 *
 * and then just add in the implementation file :
 *
\code
 Message1::registered = InternalMessage::
   register_type(MyMessageType::Message1,
                 &Message1::create)
\endcode
 *
 * Now we can easily use hour message :
\code
Message1 m(MY_ID , DESTINATION_ID);//create a new Message1 adressed to
                                 // DESTINATION_ID, signed by MY_ID
m.setFoo(42.0);
m.setBar(3.14159265);
m.send_me();//message sended over the network.
\endcode
 *
 */

#define IS_A_MESSAGE(MessageClassName, MessageEnumName)\
  private :\
    static bool registered;\
  protected :\
  static InternalMessage* create(){ return new MessageClassName();}\
  InternalMessage* clone(){return new MessageClassName(*this);}\
  public :\
  static MessageClassName cast(Message &m){\
    return MessageClassName(*dynamic_cast<MessageClassName*>\
                                    (m.get_internal_message()));\
  }\
  virtual int size(){return 0;}\
  virtual MessageType type() {return static_cast<MessageType>(MessageEnumName);}\
  MessageClassName(int idSend, int idRecv)\
    : InternalMessage(idSend,idRecv){\
  }\
  MessageClassName(){};


#define ALL_RECIPIENTS (-1)
#define MAX_TYPES 256


/**
 * defines a message that can be send and received.
 */

class InternalMessage;
// Pointer to a function which creates a message.

/**
 * Pointer to a Message creator.
 *
 */
typedef InternalMessage* (*MESSAGE_CREATOR)();

/**
 * Definition of the interface of a module which can emit and receive.
 * This class is statically used to launch and receive message in your program.
 * For this purpose, you should create the appropriate EmmiterReceiver according
 * to the type of webots component you are using ;
 * - for controller/supervisor : EmitterReceiverRobot
 * - for physics plugins : EmitterReceiverPhysics
 *
 * \attention The EmitterReceiverPhysics still have some limitation.
 * This is principaly an abstraction of the webots class for sending and getting
 * the data through sendPacket() and getData(). End users of the library should
 * not manipulate it directly, but rather use the initialization methods of the
 * Message class :
\code
  Message::set_emitter_receiver(EmitterReceiverRobot::create(this));
  Message::receive_enable(timeStep_);
\endcode
 * Then all sending and receiving of the data are managed by the EmitterReceiver
 * classes.
 */
class EmitterReceiver {

  /**
   * Type of the EmitterReceiver
   */
  enum Type{
    ROBOT,//!< ROBOT for webots controller / supervisor programs, see EmitterReceiverRobot.
    PHYSIC//!< PHYSIC for Physics plugins, see EmitterReceiverPhysics.
  };
public:
  //Emitter methods
  /**
   * Send a packet of data.
   *
   * @param data the array of data
   * @param size its size.
   * @return the size sended.
   */
  virtual int sendPacket(const void *data, int size) = 0;

  //Receiver methods
  /**
   * Enables the update of the webots::Receiver* and webots::Emitter*;
   * @param ms the timestep to update, must be a multiple of the timestep of the
   *  simulation, see webots reference.
   */
  virtual void enable(unsigned short ms) = 0;

  /**
   * Gets the length of the queue
   * @return the lenth of the queue.
   */
  virtual int getQueueLength() = 0;
  /**
   * get the next packet.
   */
  virtual void nextPacket() = 0;
  /**
   * return the size of the current packet.
   * @return size of the current packet.
   */
  virtual int getDataSize() = 0;
  /**
   * Returns a pointer to the data.
   * \return a pointer to the data
   */
  virtual const void *getData() = 0;
  /**
   * Destructor.
   *
   */
  virtual ~EmitterReceiver(){}
};

/**
 * \brief Class to handle all kind of messages.
 *
 * This is the main class to manipulate. Its give IO acces to message rather
 * than raw data.
 *
 * In order to function properly, the EmmiterReceiver should be set as soon as
 * possible. For example for a webots::Robot* controller you should use :
\code
  Message::set_emitter_receiver(EmitterReceiverRobot::create(this));
  Message::receive_enable(timeStep_);
\endcode
 * Then you can access to the message through the static methods
 * waiting_messages() and receive_message() :
\code
while(Message::waiting_message()){
  Message m(Message::receive_message());
  //do stuff
}
\endcode
 * One can see if the message is adressed to us , using the id identification
 * protocol with is_valid() and am_I_receiver(). Then we can access to the type
 * of message and then to the data it contains see IS_A_MESSAGE macro definition :
\code
if(m.is_valid()&&m.am_I_receiver(MY_ID)){
  switch(m.type()){
  case DUMMY_MESSAGE :{
    DummyMessage mm=DummyMessage::cast(m);
    //do stuff with our DummyMessage ....
    break;
  }
  default :
    cerr<<"Unhandled Message type"<<type<<endl;
  }
}
\endcode
 *
 *
 * \author Mikaël Mayer
 * \date 04/11/2008
 */
class Message {
protected:
  /// Internal message.
  InternalMessage* m;

public:
  /// Internal constructor
  Message(InternalMessage*);



  /// Constructor
  Message();

  /// Destructor
  ~Message();

  /// Copy constructor
  Message(const Message &message);

  /// Affectation operator.
  Message& operator=(const Message &message);

  /// Retrieves the internal message.
  InternalMessage* get_internal_message();

  /// @name Functions to parse incoming messages
  //@{

  /**
   * Returns if the current message is valid.
   * Should always be called before using the message.
   * \return \c true if valid.
   */
  bool is_valid();

  /// Returns the type of the message
  MessageType type();

  /// Return the id of the sender.
  int get_id_sender();

  /// Return the id of the receiver.
  int get_id_receiver();

  /**
   *  Returns \c true if this message is destinated to entity \c id.
   *  \param id The id of the module.
   *  \return \c true if the message is correctly destinated to entity.
   */
  bool am_I_receiver(int id);
  //@}

  /// @name Functions to handle reception of messages.
  //@{

  /**
   * Sets up the EmitterReceiver module. You should call it before any call to
   * other static methods of the class.
   * \sa receive_enable();
   * @param er a pointer to the EmitterReceiver.
   */
  static void set_emitter_receiver(EmitterReceiver *er);

  /**
   * Sets the period at which message should be getted from the webots::Receiver*
   * object.
   * @param step a multiple of the basic timestep of the simulation.
   */
  static void receive_enable(int step);

  /**
   * Receives a new waiting message from the receiver queue.
   *
   * \attention check if there is a Message in the queue using waiting_message().
   * @return the new Message.
   */
  static Message receive_message();

  /**
   * Receives a message from a char* and a int (if you transit with another
   * things that webots, like TCP.
   * @param data
   * @param size
   * @return
   */
  static Message receive_message(const unsigned char* data, int size);

  /**
   * gets the message as an array of data.
   * Array is created and it is up to you to manage the memory allocated
   * (just destroy it when over)
   *  @param data the pointer to the data
   *  @return the size of the data, in sizeof(unsigned char).
   */

  int get_serialized(unsigned char* data);



  /**
   * Indicate if there are any messages waiting to be picked up.
   *
   * \return \c true if there is any message to be read.
   */
  static bool waiting_messages();
  //@}
};

/**
 * \brief Abstract Class InternalMessage.
 *
 *   Skeletton for the several messages that the different modules can
 *   send each other. Each sender/receiver must have an ID in order to
 *   communicate. Each emitter/receiver of the webots robots should
 *   have the same channel. Then when an entity sends a message, each
 *   receiver get them in their queue, and look at if the message is
 *   adressed to itself.
 *
 *   This class is abstract, so you have to create your own messages in
 *   order to create new type of messages.
 *
 * \author Mikaël Mayer
 * \date 14/10/2008
 */
class InternalMessage {

  // I know it's a bad design but I making a quick patch.
  friend class EmitterReceiverPhysics;

private:
  static std::list<MessageType>* registeredTypes_;

  /// ID of the sender.
  int id_send;

  /// ID of the receiver. -1 means for all.
  int id_recv;

  /*!Return the emitter of the module.  Caution : the emitter use to
   *send message have to be called "emitter"
   *
  static Emitter* retrieve_emitter();
   *!Return the receiver of the module. Caution : the receiver use to
   * receive message have to be named "receiver"
   *
  static Receiver* retrieve_receiver();*/

  /*! All registered types of messages.
   * A message cannot be sent if its create function has not been registered.
   * \sa register_type
   */
  static MESSAGE_CREATOR registered_types[MAX_TYPES];

  /// The current number of registered types of messages
  static int num_registered_types;

protected:

  /// @name Function to register an inherited type of message.
  /**
   * Registers that a particular message exists.
   * An InternalMessage cannot be sent if its class has not been registered.
   * \param type The type of the message, coded on an 8-bit character.
   * \param registration A pointer to a static function that can create an
   *        InternalMessage
   * \return \c true if the MessageType has been succesfully registered.
   */
  static bool register_type( MessageType type, MESSAGE_CREATOR registration);

private:
  /**
   * Gets the creator of the class
   * \sa register_type
   */
  static MESSAGE_CREATOR get_message_creator(unsigned char type);

  /**
   * Returns a boolean indicating if the class of the message is registered.
   * \sa register_type
   */
  bool is_registered();

protected:
  /// @name Constructors
  //@{
  InternalMessage() {}

  /// Construct a message from a sender to a receiver
  InternalMessage(int id_send, int id_recv);
  //@}

  /// Destructor
  virtual ~InternalMessage() {}

  /// Gets message from data and size
  static InternalMessage* get_message_from_data(const unsigned char *data, int size);

public:
  /// @name Functions to parse incoming messages
  //@{
  /// Return the id of the sender.
  int get_id_sender();

  /// Return the id of the receiver.
  int get_id_receiver();

  /**
   * Return \c true if this message is destinated to entity \c id.
   * \param id The id of the module.
   * \return \c true if this message is destinated to entity \c id.
   */
  bool am_I_receiver(int id);
  //@}

  void send_me(){
    send_message(this);
  }

protected:
  static EmitterReceiver* emitter_receiver;

  /// @name Functions to send/receive messages.
  //@{
  /*! Sends a message.
   * \param message The message to send.
   */
  static void send_message(InternalMessage* message);


  /// Receives a message from the receiver queue.
  static InternalMessage* receive_message();

  ///Receive messaeg directly from data
  static InternalMessage* receive_message(const unsigned char* data, int size);

  ///  Returns true if there are any message in the queue.
  static bool waiting_messages();
  //@}

protected:
  /// @name Un/serialization functions
  //@{
  void serialize_complete(unsigned char** data, int *size);
  void unserialize_complete(const unsigned char* data, int size);
  int  size_complete();

  /// Standard serialization
  /*unsigned char* add_data(unsigned char* &data,
                          const void* field,
                          int field_size);
  const unsigned char* get_data(const unsigned char* &data,
                                void* field,
                                int field_size);*/
  //@}

  /// @name Unserialization methods for the children.
  //@{
  Serializer& serialize_data(Serializer& s);
  Serializer& unserialize_data(Serializer& s);
  int         size_data();
  //@}

  /*! @name Abstract methods
   */
  //@{
  /**
   *  Custom serialize function
   * The subclass should exactly write at adress on_data a block of size() chars
   * containing the serialized data.
   * It should then return the pointer to the next available serialization slot,
   * at address on_data + size()
   *
   * @param s a reference to the Serializer to use.
   * @return the reference to the same Serializer.
   */
  virtual Serializer&       serialize(Serializer& s) = 0;
  /**
   * Custom unserialize function
   * A block of size() chars at adress from_data is available for the subclass
   * to get its data from.
   * It should also return a pointer to the next data to unserialize,
   * at adress from_data + size()
   * @param s a reference to the Serializer to use.
   * @return the reference to the same Serializer.
   */
  virtual Serializer& unserialize(Serializer& s) = 0;
  /// Returns the exact size used by the serialize and unserialize method.
  virtual int                  size() = 0;
  /// Cloning function
  virtual InternalMessage*     clone()=0;

  virtual MessageType type() = 0;
  //@}

  friend class Message;
};
//@}

// Inline functions.

inline InternalMessage* Message::get_internal_message() {
  return m;
}

inline bool Message::is_valid() {
  return m!=NULL;
}

inline bool Message::waiting_messages() {
  return InternalMessage::waiting_messages();
}

inline void Message::receive_enable(int step) {
  InternalMessage::emitter_receiver->enable(step);
}

#endif //MESSAGE_HPP
