/*
 * Copyright (c) 2008-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 *  \file serializer.hpp
 *  \author TULEU Alexandre
 *  \date  10/12/2008
 *  \brief This file is the headers of the Serializer class.
 */

#ifndef __SERIALIZER_H_
#define __SERIALIZER_H_

#define STARTING_SIZE_OF_BLOCK 10

/*! class Serializer
 * \brief Implements a memory un/serializer
 *
 * This class implements a memory serializer. Its purpose is to
 * serialize data in a unsigned char array, throught a stl cout-like
 * interface. It ensures that the order in memory is preserved. It
 * optimizes the amount of memory used. Note that all array used are
 * copied, so you still have to manage the memory you pass/get from
 * the object of this class.
 *
 *You can create a Serializer and serialize data like this :
 * \code
 *
 * int a=1
 * int b=1138;
 *
 * float pi= 3.14159;
 * Serializer s;
 *
 * s<<a<<b<<pi;
 *
 * \endcode
 *
 * Then you simply unserialize data like this :
 *
 * \code
 * int aa,bb,cc,dd;
 * float ppi;
 *
 * s>>aa>>bb>>cc>>ppi;
 * cout<<aa<<" "<<bb<<" "<<cc<<" "<<ppi<<endl;
 * \endcode
 * this will display on the standart output "1 1138 3.14159".
 *
 * Warning, if you do not unserialize the data in the same order, it
 * will absolutely not work. But, if you ask to unserialize
 * data that was not serialized, it will left you with NULL
 * object. For example, with the code upper, if we do now :
 *
 * \code s>>dd; \endcode
 * the value of dd will simply be 0.
 *
 * You could have acces to the serialized data through the methods
 * get_size() and get_an_copy_of_data(). You could retrieve the data
 * by creating a new Serializer object, filled with serialized data
 * through the dedidaceted constructor. Notice that when you get an
 * access to the data, you have acces to a brand new copy of the data
 * and that you need to manage the memory.
 *
 *  \code 
 * int a =13; 
 * Serializer s; 
 * s<<a;
 *
 * Serializer ss=Serializer(s.data(), s.size());
 *
 * int aa;
 * ss>>aa;
 * \endcode
 *
 * aa will have the value 13. This is useful to throw data throught a
 * webots message, for example.
 *
 * \ingroup Message
 */


class Serializer {

 public :
  ///@name Constructor and destructor
  //@{
  /*! The standard constructor*/
  Serializer();
  /**
   *  Creates the serializer from a String of serialized data. Note
   *  the new serializer will have his serialize_pointer at the end
   *  (ready to append new data), and his unserialized_pointer at the
   *  beginning (ready to unserialized data). The array data is
   *  recopied in memory, so you still manage the memory of the
   *  argument.
   *  @param data the string to use.
   *  @param size the size of the string.
   */
  Serializer(const unsigned char* data, int size);
  /*! The standard destructor */
  ~Serializer();
  //@}

  /// Gets the length of the serialized data
  int get_size();

  /**
   * Gets an copy of the data as a string, (serialized data). It is
   *  up to you to handle the memory allocated.
   *  \return the serialized data in a c-string array
   */
  unsigned char* get_copy_of_data();

  /**
   *  Set the serialized data of the Serializer. It will destroy previous
   *  Serialized data
   *  @param data_ the string to use.
   *  @param size_ the size of the string.
   */
  void set_serialized_data(const unsigned char* data_,int size_);


  /**
   * @name Serializer operator
   * Serialization operation.
   * \todo I was young and innocent when I wrote this ... TEMPLATE IT !!!!
   */
  //@{
  ///append a char to the data contained in the serializer.
  friend Serializer& operator<< (Serializer& s, const char v);
  ///append a double to the data contained in the serializer.
  friend Serializer& operator<< (Serializer& s, const double v);
  ///append a float to the data contained in the serializer.
  friend Serializer& operator<< (Serializer& s, const float v);
  ///append an int to the data contained in the serializer.
  friend Serializer& operator<< (Serializer& s, const int v);
  ///append an bool to the data contained in the serializer.
  friend Serializer& operator<< (Serializer& s, const bool v);
  //@}

  ///@name unserializer operator
  //@{
  /**
   * get the next char in the data. WARNING : the order must be the
   * same than serialization. In other way, you will just get back
   * meaningless data.
   * @param s the Serializer to use
   * @param v a reference to store the extracted character.
   * @return the used Serializer
   */
  friend Serializer& operator>> (Serializer& s, char& v);
  /**
   * get the next double in the data. WARNING : the order must be the
   * same than serialization. In other way, you will just get back
   * meaningless data.
   * @param s the Serializer to use.
   * @param v a reference to store the extracted double.
   * @return the used Serializer
   *
   *
   */
  friend Serializer& operator>> (Serializer& s, double& v);
  /**
   * get the next float in the data. WARNING : the order must be the
   * same than serialization. In other way, you will just get back
   * meaningless data.
   * @param s the Serializer to use
   * @param v a reference to store the extracted float.
   * @return the used Serializer
   *
   *
   */
  friend Serializer& operator>> (Serializer& s, float& v);
  /**
   * get the next int in the data. WARNING : the order must be the
   * same than serialization. In other way, you will just get back
   * meaningless data.
   * @param s the Serializer to use
   * @param v a reference to store the extracted int.
   * @return the used Serializer
   *
   *
   */
  friend Serializer& operator>> (Serializer& s, int& v);
  /**
   * get the next int in the data. WARNING : the order must be the
   * same than serialization. In other way, you will just get back
   * meaningless data.
   * @param s the Serializer to use
   * @param v a reference to store the extracted bool.
   * @return the used Serializer
   */
  friend Serializer& operator>> (Serializer& s, bool& v);
  //@}

private :
  unsigned char* data;
  int size;

  int serialize_pointer;
  int unserialize_pointer;

  //just create a bigger block
  void create_a_bigger_block();
  //generic append method, for primitive type
  void generic_add(const void* value, int size_value);
  //generic remove method, for primitive type
  void generic_remove(void* value, int size_value);
};

Serializer& operator<< (Serializer& s, const char v);
Serializer& operator<< (Serializer& s, const double v);
Serializer& operator<< (Serializer& s, const float v);
Serializer& operator<< (Serializer& s, const int v);
Serializer& operator<< (Serializer& s, const bool v);

Serializer& operator>> (Serializer& s, char& v);
Serializer& operator>> (Serializer& s, double& v);
Serializer& operator>> (Serializer& s, float& v);
Serializer& operator>> (Serializer& s, int& v);
Serializer& operator>> (Serializer& s, bool& v);
#endif // __SERIALIZER_H_
