/*
 * Copyright (c) 2009-2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Actuator.h
 * Private headers of the Actuator base class. You should rather include
 * Actuators.h
 *
 * \date 26 nov. 2009
 * \author Alexandre Tuleu
 */

#ifndef LIBAMARSI_QUAD_ACTUATOR_H_
#define LIBAMARSI_QUAD_ACTUATOR_H_


#include <webots/Servo.hpp>
#include <set>
namespace amarsi {
	/**
	 *
	 * \ingroup user
	 *
	 * Represents an actuator of the robot. It is related to a webots::Servo* which
	 * is used to act on the robot. This class has been introduce to model the
	 * Cheetah's actuator more accurately. Its also add some automatic computation
	 * of the power consumption.
	 *
	 * Some of the model need to be updated at each simulatiom step, evn if there
	 * is no changes of the user order. For this purpose, the static method
	 * updateAllActuatorStates() must be called at each timestep. This is
	 * automatically performed by the PimpController.
	 *
	 */
	class Actuator {
	public:
		enum Mode{
			POSITION_CONTROLLED,
			TORQUE_CONTROLLED
		};


		Mode mode() const;

		virtual void setMode(Mode m);
		/**
		 * Update all the state of the current program Actuator (s).
		 */
		static void updateAllStates();
		/**
		 * Send all the commands of the current program Actuator (s).
		 */
		static void sendAllCommands();

		/**
		 * Resets all the actuators.
		 */
		static void resetAllActuators();
		/**
		 * Standard Constructor
		 * @param servo the webots::Servo* to use with the actuator
		 * @param timeStep the timeStep at which the actuator will be updated
		 * @return a new actuator
		 */
		Actuator();
		virtual ~Actuator();
		/**
		 * Returns the modelized position of the actuator. Could be different of the
		 * webots field.
		 * @return the current modelized position (in radians).
		 */
		virtual double position() const=0;
		/**
		 * Returns the mechanical power added to the system for the current step.
		 * @return the mechanical power added to the system for the current step.
		 */
		virtual double power() const=0;
		/**
		 * Returns the output torque of the actuator.
		 * @return
		 */
		virtual double torque() const=0;
		/**
		 * Current velocity of the servo.
		 * @return
		 */
		virtual double velocity() const=0;

		/**
		 * Sets the new command of the servo for the next timestep.
		 */
		void setCommand(double command){
			d_desiredPosition =  command;
		}

		/**
		 * Returns the command that will be used for the next timestep.
		 */
		double command() const{
			return d_desiredPosition;
		}

		/**
		 * Sets the new order for the timestep.
		 */
		void setOrder(double desiredPosition){
			d_desiredPosition=desiredPosition;
		}

		double order() const{
			return d_desiredPosition;
		}
	protected:
		/**
		 * Updates the state of the Actuator according to the simulation state and
		 * the current order. If no new order has been given, take the older value
		 * as the new order (for actuator whom state depends on the derivative of
		 * the order).
		 */
		virtual void updateState()=0;

		virtual void sendCommand()=0;

		virtual void reset()=0;
	private :

		typedef std::set<Actuator*> SetOfActuator;
		static SetOfActuator s_actuators;

		double d_desiredPosition;
		Mode d_mode;
	};

}

#endif // LIBAMARSI_ACTUATOR_H_
