/*
 * Copyright (c) 2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file KneeActuatorAmarsi.cpp
 *
 * \date 30 sept. 2009
 * \author Alexandre Tuleu
 */
#include <iostream>
#include <iomanip>
#include <cassert>
#include <cmath>

#include <libamarsi-quad/common/CheetahMessages.h>
#include <libamarsi-quad/utils/Debug.h>
#include <libamarsi-quad/utils/PreferenceReader.h>
#include <libamarsi-quad/common/WebotsExplorer.h>
#include <libamarsi-quad/controller/Clock.h>
#include "KneeActuatorAmarsi.h"

namespace amarsi{

	KneeActuatorAmarsi::KneeActuatorAmarsi(webots::Servo* actuatedServo,
	                                       webots::Servo* passiveServo,
	                                       webots::Node* passiveServoNode,
	                                       webots::Node* actuadedServoNode,
	                                       Limbs l)
	: Actuator()
	, d_passiveJoint(passiveServo)
	, d_activeJoint(actuatedServo)
	, d_passiveJointNode(passiveServoNode)
	, d_activeJointNode(actuadedServoNode)
	, d_l(l)
	, d_kneeStiffness(0.53)
	, d_kneeDamping(0.03)
	, d_maxForce(3)
	, d_maxVelocity(29.60)
	, d_maxCompliantVelocity(8.0)
	, d_x1(0.0)
	, d_x1Dot(0.0)
	, d_x1Des(0.0)
	, d_x1DesDot(0.0)
	, d_tau1(0.0)
	, d_x1DesPrev(0.0)
	, d_elasticFactor(1.0)
	, d_kp(0.5)
	, d_kd(0.02)
	, d_m(NONE){
		init(actuadedServoNode);
	}

	//  double KneeActuatorAmarsi::getLinearizedStiffness(double stiffness) const{
	//    return std::abs(model.torqueFromAngle(stiffness,
	//                                          model.maximalAmplitudeKnee())
	//                      /model.maximalAmplitudeKnee());
	//  }
	KneeActuatorAmarsi::~KneeActuatorAmarsi(){
	}

double KneeActuatorAmarsi::computeX1DesAndDampTorque(){
    double speed = (d_x1Des - d_x1DesPrev) / amarsi::Clock::timeStepInS();
    if (speed > d_maxCompliantVelocity)
      d_x1Des = Clock::timeStepInS() * d_maxCompliantVelocity + d_x1DesPrev;
    if (speed < -d_maxCompliantVelocity)
      d_x1Des = - Clock::timeStepInS() * d_maxCompliantVelocity + d_x1DesPrev;

    double computedTorque = (d_x1Des - d_x1) * d_kneeStiffness;
    if (computedTorque > d_maxForce)
      d_x1Des = d_maxForce / d_kneeStiffness + d_x1;
    if (computedTorque < -d_maxForce)
      d_x1Des = -d_maxForce / d_kneeStiffness + d_x1;

    d_x1Des = std::min(0.0, std::max(d_x1Des, d_maxPosition));

    d_x1DesDot = (d_x1Des - d_x1DesPrev) / Clock::timeStepInS();

    return std::min(d_maxDampingTorque,
                    std::max(-d_maxDampingTorque, -d_kneeDamping
                             * (d_x1Dot)));
  }

  void KneeActuatorAmarsi::computeTorqueNone(){
    double dampTorque= computeX1DesAndDampTorque();
    d_tau1 = d_kneeStiffness * (d_x1Des - d_x1) + dampTorque;
  }

  void KneeActuatorAmarsi::computeTorqueElastic(){
    double dampTorque= computeX1DesAndDampTorque();
    d_tau1 = d_kneeStiffness * (d_x1Des >= d_x1 ? 1.0 : d_elasticFactor )*(d_x1Des - d_x1) + dampTorque;
  }

  void KneeActuatorAmarsi::computeTorquePD(){
    double dampTorque= computeX1DesAndDampTorque();
    if(d_x1Des < d_x1){// in this case we use PD control
      d_tau1 = d_kp * (d_x1Des - d_x1) + d_kd * ( d_x1DesDot - d_x1Dot );
    }else{//we use elastic control
      d_tau1 = d_kneeStiffness * ( d_x1Des - d_x1 ) + dampTorque;
    }

  }

	void KneeActuatorAmarsi::updateState()
	{
		d_activeJoint.update();
		d_passiveJoint.update();
	}

	void KneeActuatorAmarsi::sendCommand(){
		//activeServoSpeed_=(servo_->getPosition()- activePreviousPosition_)/timeStep_;
		//activePreviousPosition_=(servo_->getPosition());



		d_x1Dot = d_passiveJoint.velocity();
		d_x1 = d_passiveJoint.position();
		if(mode() == Actuator::POSITION_CONTROLLED){
			d_x1DesPrev=d_x1Des;
			d_x1Des=std::min(0.0,std::max(-order(),d_maxPosition));

			switch(d_m){
			case NONE :
				computeTorqueNone();
				break;
			case ELASTIC :
				computeTorqueElastic();
				break;
			case PD :
				computeTorquePD();
				break;
			}
		}else{//Directly Torque controlled
			d_tau1 = - order();

			d_x1DesPrev = d_x1Des;
			d_x1Des = d_tau1 /d_kneeStiffness + d_x1;
			d_x1DesDot = (d_x1Des - d_x1DesPrev) / amarsi::Clock::timeStepInS();

		}

    double tmpDpTRque = std::min(d_maxDampingTorque,std::max(-d_maxDampingTorque,
                                                             -d_kneeDamping* (d_x1Dot)));
    std::cout << std::showpos << std::scientific << std::setprecision(2);
    debug_out_nh(CONTROLLER,d_m<<"x1 : "<<d_x1<<" "<<
        "x1D : "<<d_x1Des<<" "<<
        "order : "<<std::min(0.0,std::max(-order(),d_maxPosition))<<" "<<
        "x1Dt : "<<d_x1Dot<<" "<<
        "x1DDt : "<<d_x1DesDot<<" "<<
        "kStiff : "<<d_kneeStiffness<<" "<<
        "tau1 :"<<d_tau1<<" "<<
        "dpTrq :"<<tmpDpTRque<<" "
        "mxPos : "<<d_maxPosition);
#ifdef DEBUG
		double alpha1=0.7854 + d_passiveJoint.position();//negative
		double alpha2=0.7854 - d_activeJoint.position();//positive
		double relativeLength = std::sqrt((3 - 2 * std::cos( alpha1 )
		- 2 * std::cos( alpha2 )
		+ 2 * std::cos( alpha1 - alpha2 ))/
		(5 - 4 * std::cos(0.7854)));
#endif //DEBUG
//        debug_controller("Relative Length : "<<relativeLength<<std::endl);
		debug_out_nh(CONTROLLER,"rLen : "<<relativeLength<<std::endl);
		//        KneeOrder m(ID_ROBOT,ID_PHYSICS);
		//        m.setLimb(l_);
		//        m.setTorque(tau1_);
		//        m.send_me();
		//        std::cerr<<tau1_<<std::endl;
		//
		//        passiveServo_->setForce(0.000000001);
		d_passiveJoint.sendCommand(d_tau1);
		d_activeJoint.sendCommand(-d_x1);
	}

	void KneeActuatorAmarsi::initValuesFromPrefs(){
		PreferenceReader & r = PreferenceReader::instance();
    	std::string mode;
		r.setDomain("KneeServo");
		r.getValue("stiffness", d_kneeStiffness);
		r.getValue("damping", d_kneeDamping);
		r.getValue("max_torque", d_maxForce);
		r.getValue("max_speed", d_maxVelocity);
		r.getValue("max_damping_torque",d_maxDampingTorque);
		r.getValue("max_compliant_velocity",d_maxCompliantVelocity);
    r.getValue("assymetric_mode",mode);
    r.getValue("elastic_factor",d_elasticFactor);
    r.getValue("pd_kd", d_kd);
    r.getValue("pd_kp", d_kp);
    if(mode=="ELASTIC"){
      d_m=ELASTIC;
    }
    if(mode=="PD"){
      d_m=PD;
    }
	}

	void KneeActuatorAmarsi::init(webots::Node *actuatedServoNode){
		d_activeJoint.setMode(ServoDriver::POSITION_DRIVEN);
		d_passiveJoint.setMode(ServoDriver::TORQUE_DRIVEN);

		initValuesFromPrefs();
		webots::Field *activeMaxForce, *activeMaxVelocity,*activeMaxPosition;
		WebotsExplorer::getField(actuatedServoNode, "maxForce",&activeMaxForce);
		WebotsExplorer::getField(actuatedServoNode, "maxVelocity",&activeMaxVelocity);
		WebotsExplorer::getField(actuatedServoNode, "maxPosition",&activeMaxPosition);
		activeMaxForce->setSFFloat(d_maxForce);
		activeMaxVelocity->setSFFloat(d_maxVelocity);
		d_maxPosition=-activeMaxPosition->getSFFloat();
	}

	//  KneeActuatorAmarsi::LegModel::LegModel(webots::Node * leg, double deltaTibia)
	//    : deltaTibia_(std::abs(deltaTibia)){
	//    init(leg);
	//  }
	//  double KneeActuatorAmarsi::LegModel::angleFromX(double x) const{
	//    return  std::acos((x*x-delta_L_2_plus_L_2_2_)/_2_delta_L_L_2_)
	//           - initialKneeAngle_;
	//  }
	//  double KneeActuatorAmarsi::LegModel::xFromAngle(double angle) const{
	//    return std::sqrt(delta_L_2_plus_L_2_2_
	//                     + _2_delta_L_L_2_*std::cos(initialKneeAngle_-angle));
	//  }
	//  double KneeActuatorAmarsi::LegModel::torqueFromAngle(double fitness,
	//                                                       double angle) const{
	//    return A*std::sin(initialKneeAngle_-angle)*fitness*(1-x_0/xFromAngle(angle));
	//  }
	//
	////
	////        //convenience pre-computed terms
	////        double delta_L_2_plus_L_2_2_, // \delta_L^2 + L_2^2
	////               _2_delta_L_L_2_,// 2 * \delta_L * L_2
	////               x_0,//cos(\alpha_0)
	////               A;//-(L1 - \delta_L /2 )*L2*\delta_L/(L1+L3-\delta_L)
	//
	//  void KneeActuatorAmarsi::LegModel::init(webots::Node *leg){
	//    std::vector<webots::Node*> femurChildrens;
	//    assert(WebotsExplorer::getChildren(femurChildrens,leg,2));
	//
	//    webots::Field *tibiaTrans,*tibiaRotation,*amplitude;
	//    assert(WebotsExplorer::getField(&tibiaTrans,femurChildrens[0],"translation"));
	//    assert(WebotsExplorer::getField(&tibiaRotation,femurChildrens[0],"rotation"));
	//    assert(WebotsExplorer::getField(&amplitude,femurChildrens[0],"minPosition"));
	//    femurLength_=std::abs(tibiaTrans->getSFVec3f()[1]);
	//    initialKneeAngle_=3.14159265-std::abs(tibiaRotation->getSFRotation()[3]);
	//    maxAmpKneeAngle_=std::abs(amplitude->getSFFloat());
	//
	//    std::vector<webots::Node*> tibiaChildrens;
	//    std::vector<webots::Node*> ankleChildrens;
	//    assert(WebotsExplorer::getChildren(tibiaChildrens,femurChildrens[0],2));
	//    assert(WebotsExplorer::getChildren(ankleChildrens,tibiaChildrens[1],2));
	//
	//
	//
	//    webots::Field *ankleTrans,*toeTrans;
	//    assert(WebotsExplorer::getField(&ankleTrans,tibiaChildrens[1],"translation"));
	//    assert(WebotsExplorer::getField(&toeTrans,ankleChildrens[1],"translation"));
	//    tibiaLength_=std::abs(ankleTrans->getSFVec3f()[1]);
	//    footLength_=std::abs(toeTrans->getSFVec3f()[0])+deltaTibia_/2.0;
	//
	//    delta_L_2_plus_L_2_2_=deltaTibia_*deltaTibia_+tibiaLength_*tibiaLength_;
	//    _2_delta_L_L_2_=2*deltaTibia_*tibiaLength_;
	//    x_0=xFromAngle(0.0);
	//    A=-(footLength_- deltaTibia_/2.0)*tibiaLength_*deltaTibia_
	//        /(femurLength_+footLength_-deltaTibia_);
	//
	//  }
	double KneeActuatorAmarsi::position() const{
		return - d_x1Des;
	}
	double KneeActuatorAmarsi::velocity() const{
		return - d_x1DesDot;
	}
	double KneeActuatorAmarsi::torque() const{
		return - d_kneeStiffness * (d_x1Des - d_x1);
	}
	void KneeActuatorAmarsi::reset(){
		d_activeJointNode.reset();
		d_activeJoint.reset();
		d_passiveJoint.reset();
		d_passiveJointNode.reset();
		d_x1Des     = 0.0;
		d_x1Dot     = 0.0;
		d_x1Des     = 0.0;
		d_x1DesDot  = 0.0;
		d_tau1      = 0.0;
		d_x1DesPrev = 0.0;
	}



	double KneeActuatorAmarsi::power() const{
		return d_x1DesDot * d_kneeStiffness * (d_x1Des - d_x1) + d_activeJoint.power();
	}
}

