/*
 * Copyright (c) 2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file KneeActuatorAmarsi.h
 *
 * \date 30 sept. 2009
 * \author Alexandre Tuleu
 */

#ifndef LIBAMARSI_QUAD_KNEE_ACTUATOR_AMARSI_H_
#define LIBAMARSI_QUAD_KNEE_ACTUATOR_AMARSI_H_

#include "Actuator.h"
#include "ServoDriver.h"
#include "NodeResetter.h"
#include <libamarsi-quad/filtering/Rate.h>
namespace webots{
	class Node;
}

namespace amarsi {

	/**
	 * An Actuator that aims to simulate the behavior of the Cheetah's leg for our
	 * simplified model. Actually this Actuator use two webots::Servo*, one
	 * Being an almost perfect servo (instant change of position in the boundary
	 * of its characteristic) in series with a spring-damper mechanism, and one
	 * webots-simulated P-controlled servomotor.
	 *
	 * The idea is to simulate the pantographic behavior by
	 * making the P-controlled joint following the position of the other. The user
	 * should only give the order for the latter, and the KneeActuatorAmarsi
	 * handles the rest of the command.
	 *
	 * N.B. : this class use the Preference mechanism (see PreferenceReader) to handle
	 * some configuration value. All of this value are in the "KneeServo" domain
	 *   - "stiffness", specify the linear stiffness of the knee joint
	 *   - "damping", specify the damping coefficient of the knee joint
	 *   - "max_torque", specifies the maximal output torque of the servo, (for the
	 *      perfect and controlled one)
	 *   - "max_speed", specifies the maximal output velocity of the P-controlled
	 *     servo.
	 *   - "max_damping_torque" add a saturation of the knee damping force, to
	 *      correct stability problem.
	 *   - "max_compliant_velocity" specifies the maximal output velocity of the
	 *      perfect knee servo. Should be the same than the ankle servo, but is
	 *      lower to correct some instability
	 *
	 *  \ingroup internal
	 */
	class KneeActuatorAmarsi : public Actuator{
	public :
		/**
		 * Stadndard constructor
		 * @param actuatedServo the lower (ankle), P-controlled joint
		 * @param passiveServo the upper (knee), compliant joint
		 * @param actuatedServoNode pointer to the ankle node.
		 * @param timeStep integration timestep in seconds.
		 * @param l the limb of the KneeActuatorAmarsi
		 * @return
		 */
		KneeActuatorAmarsi(webots::Servo* actuatedServo,
		                   webots::Servo* passiveServo,
		                   webots::Node* passiveServoNode,
		                   webots::Node* actuatedServoNode,
		                   Limbs l);

		virtual ~KneeActuatorAmarsi();


		virtual void updateState();
		virtual void sendCommand();
		virtual void reset();
		virtual double power() const;
		virtual double position() const;
		virtual double velocity() const;
		virtual double torque() const;
		//    double getLinearizedStiffness(double stiffness) const;

	private :
		enum Mode {
			NONE,
			ELASTIC,
			PD
		};
		/**
		 * Initialize some webots fields value and values of the class.
		 * @param actuatedServoNode code of the ankle joint.
		 */
		void init( webots::Node* actuatedServoNode);
		/**
		 * Initialize the charcteristic of the servo according to the Preferences
		 * (see PreferenceReader).
		 */
		void initValuesFromPrefs();
		double computeX1DesAndDampTorque();
		void computeTorqueNone();
		void computeTorqueElastic();
		void computeTorquePD();



		ServoDriver d_passiveJoint;
		ServoDriver d_activeJoint;
		NodeResetter d_passiveJointNode;
		NodeResetter d_activeJointNode;
		Limbs d_l;
		double d_kneeStiffness;
		double d_kneeDamping;
		double d_maxForce;
		double d_maxVelocity;
		double d_maxCompliantVelocity;
		double d_maxPosition;
		double d_maxDampingTorque;
		double d_x1;
		double d_x1Dot;
		double d_x1Des;
		double d_x1DesDot;
		double d_tau1;
		double d_x1DesPrev;
		double d_elasticFactor;
		double d_kp;
		double d_kd;
		Mode d_m;
	};
}

#endif //LIBAMARSI_QUAD_KNEE_ACTUATOR_CHEETAH_H_
