/*
 * Copyright (c) 2009-2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Vector-inl.h
 *
 *  \date Jun 17, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_VECTOR_INL_H_
#define LIBAMARSI_QUAD_VECTOR_INL_H_

#include <cstdlib>
#include <cstring>

#define VECTOR_FOR_ALL_I for(uint i=0;i<N;++i)
#define VECTOR_FOR(start, end) for(uint i=start;i<end;++i)

//==============================================================================
// Constructors and destructors
//==============================================================================
template<uint N> Vector<N>::Vector(bool zero) {
  if (zero)
    setZero();
}

template <> inline Vector<2>::Vector(bool zero){
  if(!zero)
    return;
  _[0]=0.0;
  _[1]=0.0;
}

template <> inline Vector<3>::Vector(bool zero){
  if(!zero)
    return;
  _[0]=0.0;
  _[1]=0.0;
  _[2]=0.0;
}

template <> inline Vector<4>::Vector(bool zero){
  if(!zero)
    return;
  _[0]=0.0;
  _[1]=0.0;
  _[2]=0.0;
  _[3]=0.0;
}

template <uint N>template <uint NO> Vector<N>::Vector(const Vector<NO> & o){
  uint min=N<NO?N:NO;
  for(uint i=0;i<min;++i)
    _[i]=o[i];
  setZero(NO);
}


template <uint N> inline Vector<N>::Vector( const double *a){
  memcpy( &_[0] , &a[0] ,N*sizeof(double));
//  VECTOR_FOR_ALL_I
//    _[i]=a[i];

}


template <uint N> inline Vector<N>::Vector(const Vector<N> & src ){
  memcpy(&_[0] , &src._[0], N*sizeof(double));

//  VECTOR_FOR_ALL_I
//    _[i]=src._[i];
}



template <uint N> Vector<N>::~Vector(){
}

//==============================================================================
// Static Members
//==============================================================================


template <uint N> const Vector<N> Vector<N>::ZERO(true);

template <uint N> Vector<N> Vector<N>::E(uint i){
  Vector<N> res(true);//Zero vector
  if(i<N)
    res._[i]=1.0;
  return res;
}

//==============================================================================
// Access Operators
//==============================================================================

template <uint N> inline double Vector<N>::operator[](uint i) const{
  if(i<N)
    return _[i];
  return 0.0;/// \todo should throw an exception here !!!!
}
template <uint N> inline double & Vector<N>::operator[](uint i){
  if(i<N)
    return _[i];/// \todo should thrw an exception here !!!!
  return _[0];
}

//==============================================================================
// Normalization methods
//==============================================================================

template <uint N> inline double Vector<N>::norm() const{
  return std::sqrt(norm2());
}

template <uint N> double Vector<N>::norm2() const{
  double res(0.0);
  VECTOR_FOR_ALL_I
    res+=_[i]*_[i];
  return res;
}

template <> inline double Vector<2>::norm2() const{
  return _[0]*_[0] + _[1] * _[1];
}

template <> inline double Vector<3>::norm2() const{
  return _[0]*_[0] + _[1] * _[1] + _[2] * _[2];
}

template <> inline double Vector<4>::norm2() const{
  return _[0]*_[0] + _[1] * _[1] + _[2] * _[2] + _[3] * _[3];
}

template <uint N> double Vector<N>::normalize(){
  double n=norm();
  if(n<BASIC_MATH_ZERO_THRESHOLD)
    setZero();
  else
    *this/=n;
  return n;
}

template <uint N> Vector<N> Vector<N>::normalized() const{
  Vector<N> res(*this);
  return res.normalize();
}

//==============================================================================
// State Methods
//==============================================================================

template <uint N> bool Vector<N>::isNan() const{
  bool res=false;
  for(uint i=0;i<N && !res; ++i)
    res=_[i]!=_[i];//true if isNaN
  return res;
}

template <> inline bool Vector<2>::isNan() const{
  return _[0] != _[0] || _[1] || _[1] ;
}

template <> inline bool Vector<3>::isNan() const{
  return _[0] != _[0] || _[1] != _[1] || _[2] != _[2] ;
}

template <> inline bool Vector<4>::isNan() const{
  return _[0] != _[0] || _[1] != _[1] || _[2] != _[2] || _[3] != _[3] ;
}

//==============================================================================
// Operation methods
//==============================================================================

template <uint N> void Vector<N>::add(const Vector<N> & other,
                                      Vector<N> & res) const{
  VECTOR_FOR_ALL_I
    res._[i]=_[i]+other._[i];
}

template <> inline void Vector<2>::add(const Vector<2> & other,
                                       Vector<2> & res) const{
  res._[0]=_[0]+other._[0];
  res._[1]=_[1]+other._[1];
}

template <> inline void Vector<3>::add(const Vector<3> & other,
                                       Vector<3> & res) const{
  res._[0]=_[0]+other._[0];
  res._[1]=_[1]+other._[1];
  res._[2]=_[2]+other._[2];
}

template <> inline void Vector<4>::add(const Vector<4> & other,
                                       Vector<4> & res) const{
  res._[0]=_[0]+other._[0];
  res._[1]=_[1]+other._[1];
  res._[2]=_[2]+other._[2];
  res._[3]=_[3]+other._[3];
}

template <uint N> void Vector<N>::add(double a, Vector<N> & res) const{
  VECTOR_FOR_ALL_I
    res._[i]=_[i]+a;
}

template <> inline void Vector<2>::add(double a, Vector<2> & res) const{
  res._[0]=_[0]+a;
  res._[1]=_[1]+a;
}


template <> inline void Vector<3>::add(double a, Vector<3> & res) const{
  res._[0]=_[0]+a;
  res._[1]=_[1]+a;
  res._[2]=_[2]+a;
}

template <> inline void Vector<4>::add(double a, Vector<4> & res) const{
  res._[0]=_[0]+a;
  res._[1]=_[1]+a;
  res._[2]=_[2]+a;
  res._[3]=_[3]+a;
}

template<uint N>
void Vector<N>::substract(const Vector<N> & other, Vector<N> & result) const {
  VECTOR_FOR_ALL_I
    result[i] = _[i] - other._[i];
}

template<>
inline void Vector<2>::substract(const Vector<2> & other, Vector<2> & result) const {
  result[0] = _[0] - other._[0];
  result[1] = _[1] - other._[1];
}

template<>
inline void Vector<3>::substract(const Vector<3> & other, Vector<3> & result) const {
  result[0] = _[0] - other._[0];
  result[1] = _[1] - other._[1];
  result[2] = _[2] - other._[2];
}

template<>
inline void Vector<4>::substract(const Vector<4> & other, Vector<4> & result) const {
  result[0] = _[0] - other._[0];
  result[1] = _[1] - other._[1];
  result[2] = _[2] - other._[2];
  result[3] = _[3] - other._[3];
}

template <uint N> void Vector<N>::substract(double a, Vector<N> & res) const{
  VECTOR_FOR_ALL_I
    res._[i]=_[i]-a;
}

template<>
inline void Vector<2>::substract(double a, Vector<2> & result) const {
  result[0] = _[0] - a;
  result[1] = _[1] - a;
}

template<>
inline void Vector<3>::substract(double a, Vector<3> & result) const {
  result[0] = _[0] - a;
  result[1] = _[1] - a;
  result[2] = _[2] - a;
}

template<>
inline void Vector<4>::substract(double a, Vector<4> & result) const {
  result[0] = _[0] - a;
  result[1] = _[1] - a;
  result[2] = _[2] - a;
  result[3] = _[3] - a;
}



template<uint N>
void Vector<N>::prodScal(const Vector<N> & other, double & result) const {
  result = 0.0;
  VECTOR_FOR_ALL_I
    result += _[i] * other._[i];
}

template<>
inline void Vector<2>::prodScal(const Vector<2> & other, double & result) const {
  result = 0.0;
  result += _[0] * other._[0];
  result += _[1] * other._[1];
}

template<>
inline void Vector<3>::prodScal(const Vector<3> & other, double & result) const {
  result = 0.0;
  result += _[0] * other._[0];
  result += _[1] * other._[1];
  result += _[2] * other._[2];
}

template<>
inline void Vector<4>::prodScal(const Vector<4> & other, double & result) const {
  result = 0.0;
  result += _[0] * other._[0];
  result += _[1] * other._[1];
  result += _[2] * other._[2];
  result += _[3] * other._[3];
}

template <uint N> void Vector<N>::mult(double a, Vector<N> & res) const{
  VECTOR_FOR_ALL_I
    res._[i]=a*_[i];
}

template <> inline void Vector<2>::mult(double a, Vector<2> & res) const{
  res._[0]=a*_[0];
  res._[1]=a*_[1];
}

template <> inline void Vector<3>::mult(double a, Vector<3> & res) const{
  res._[0]=a*_[0];
  res._[1]=a*_[1];
  res._[2]=a*_[2];
}

template <> inline void Vector<4>::mult(double a, Vector<4> & res) const{
  res._[0]=a*_[0];
  res._[1]=a*_[1];
  res._[2]=a*_[2];
  res._[3]=a*_[3];
}

//==============================================================================
// Members operators
//==============================================================================

template <uint N> inline void Vector<N>::operator+=(const Vector<N> & other){
  this->add(other,*this);
}

template <uint N> inline void Vector<N>::operator-=(const Vector<N> & other){
  this->substract(other,*this);
}

template <uint N> inline void Vector<N>::operator+=(double a){
  this->add(a,*this);
}

template <uint N> inline void Vector<N>::operator-=(double a){
  this->substract(a,*this);
}

template <uint N> inline void Vector<N>::operator*=(double a){
  this->mult(a,*this);
}

template <uint N> inline void Vector<N>::operator/=(double a){
  this->mult(1.0/a,*this);
}

template <uint N> bool Vector<N>::operator==(const Vector<N> &other) const{
  bool result=true;
  for(uint i=0;result && i<N; ++i)
    if(std::abs((_[i]-other[i])*2/(_[i]+other[i]))>0.000001)//the value are too far away
      result=false;
  return result;
}

template <uint N> bool Vector<N>::operator!=(const Vector<N> & other) const{
  bool result=false;
  for(uint i=0;!result && i<N;++i)
    if(std::abs((_[i]-other[i])*2/(_[i]+other[i]))>0.000001)
      result=true;
  return result;
}

//==============================================================================
// External operators
//==============================================================================


template<uint NO> std::ostream & operator<<(std::ostream & out,
                                            const Vector<NO> &a) {
  out << "( ";
  VECTOR_FOR(0,NO-1)
    out << a._[i] << " , ";
  if (NO)
    out << a._[NO - 1];
  return out << " )";
}

template<uint NO>
inline double operator*(const Vector<NO> & a, const Vector<NO> &b) {
  double res;
  a.prodScal(b, res);
  return res;
}

template<uint NO>
inline  Vector<NO> operator*(const Vector<NO> & a,
                   double b){
  Vector<NO> res;
  a.mult(b,res);
  return res;
}

template<uint NO>
inline   Vector<NO> operator*(double a, const Vector<NO> & b){
  Vector<NO> res;
  b.mult(a,res);
  return res;
}


template<uint NO> inline  Vector<NO> operator+(const Vector<NO> &a,
                                              const Vector<NO> &b) {
  Vector<NO> res(false);
  a.add(b, res);
  return res;
}

template<uint NO> inline  Vector<NO> operator-(const Vector<NO> &a,
                                              const Vector<NO> &b) {
  Vector<NO> res(false);
  a.substract(b, res);
  return res;
}

//==============================================================================
// Protected members
//==============================================================================

template <uint N> void Vector<N>::setZero(uint start){
  for(uint i=start;i<N;++i)
    _[i]=0.0;
}


#endif // LIBAMARSI_QUAD_VECTOR_INL_H_
