/*
 * Copyright (c) 2009-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Vector3.h
 *
 *  \date May 7, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_VECTOR_H_
#define LIBAMARSI_VECTOR_H_
#include "global.h"
#include <ostream>
#include <cmath>

/**
 * Base class for Vector of N dimension, using templates. Basic arithmetic
 * operation on vectors are implemeted.
 *
 * \ingroup internal
 *
 * \todo comparison is badly performed, should do a nice interface with upls, and
 * things like in google code (BSD License, so go for it)....
 */

template<uint N>
class Vector {
	template<uint NO>
	friend std::ostream & operator<<(std::ostream & out, const Vector<NO> & a);
public:

	///@name Constructors and Destructor
	//@{
	/**
	 * Creates a new vector
	 * @param zero if \c true, all coocrdiantes are initialized to zero.0
	 */
	explicit Vector(bool zero = true);

	/**
	 * Copy constructor from a Vector of different dimension. If the dimension (NO)
	 *  of the source vector is smaller, \f$ \left [ NO +1 , .. ,  N \right ]\f$
	 *  coordinates of the created Vector will be zero. On the other side if the
	 *  source dimension is bigger, only the first N coordinates of the source
	 *  Vector will be copied.
	 * @param o the source Vector.
	 */
	template<uint NO> explicit Vector(const Vector<NO> & o);

	/**
	 * Creates the Vector form a raw C array. \warning This constructor assumes
	 * that the pointed array as a sufficient N dimension.
	 * @param a pointer to a N sized raw C double array.
	 */
	explicit Vector(const double *a);
	/**
	 * Copy constructor.
	 * @param src
	 */
	Vector(const Vector<N> & src);
	virtual ~Vector();
	//@}

	/// @name Predefined Vectors
	//@{
  /**
   * The zero vector.
   */
	static const Vector<N> ZERO;
	/**
	 * The canonical base Vector of coordinate i.
	 * @param i the coordinate of the base Vector which is non-zero.
	 * \return the wanted canonical Vector<N>
	 */
	static Vector<N> E(uint i);
	//@}

	///@name Coordinate accessor methods
	//@
	/**
	 * Read/Write accessor method.
	 * @param i the wanted coordinate
	 * @return a refernece to the coordinate
	 */
	double & operator[](uint i);
	/**
	 * Read- only accessor method.
	 * @param i the coordinate to read
	 * @return value of the coordinate.
	 */
	double operator[](uint i) const;

	/**
	 * Read-only accessor.
	 * @return the vector as an raw-C array.
	 */
	const double* toArray() const{
	  return _;
	}
	//@}
	///@name normalization methods
	//@{
	/**
	 * Returns the normal of the Vector.
	 * @return the normal of the Vector
	 */
	double norm() const;
	/**
	 * Returns the square of the normal.
	 * @return the square of the normal.
	 */
	double norm2() const;
	/**
	 * Normalizes the vector. If the Vector is too small, it's considered as the
	 * Zero Vector
	 * @return the normal of the vetor BEFORE normalization.
	 */
	double normalize();
	Vector<N> normalized() const;
	//@}

	/**
	 * Test if the Vector is Not A Number, i.e. one of its cooridnate is Not a
	 * Number.
	 * @return \c true if NaN.
	 */
	bool isNan() const;

  /**
   * @name Operation Methods
   *
   * All of this methods are build on the same principle : the first argument is
   * the Vector/scalar to use with the current Vector, and the second is Vector
   * in which the result will be stored.
   */

	//@{
	/**
	 * Adds two vectors.
	 */
	void add(const Vector& other, Vector& result) const;
	/**
	 * Adds a scalar to each coordinates.
	 */
	void add(double a, Vector& result) const;
	/**
	 * Substracts other Vector to current one.
	 */
	void substract(const Vector & other, Vector & result) const;
	/**
	 * Substracts a scalar to each coordinates.
	 */
	void substract(double a, Vector & result) const;
	/**
	 * Performs the scalar product of the two Vectors.
	 */
	void prodScal(const Vector& other, double & result) const;
	/**
	 * Multiplies all coordinates by a scalar.
	 */
	void mult(double a, Vector& result) const;
	//@}

	///@name assign by operation operators
	//@{
	void operator+=(const Vector<N> & other);
	void operator-=(const Vector<N> & other);
	void operator+=(double a);
	void operator-=(double a);
	void operator*=(double a);
	void operator/=(double a);
	//@}

	/**
	 * @name Comparison Operators
	 * \todo Uses strict comparison operator, should use ULPS (see goggle test
	 * code and Documentation)
	 */
	//@{
  bool operator==(const Vector<N> & other) const;
	bool operator!=(const Vector<N> & other) const;
	//@}

	/**
	 * @name Operators
	 */
	//@{
	template<uint NO> friend double operator*(const Vector<NO> & a,
	                                          const Vector<NO> &b);
	template<uint NO> friend Vector<NO> operator*(const Vector<NO> & a,
	                                          double b);
	template<uint NO> friend Vector<NO> operator*(double a,const Vector<NO> & b);

	template<uint NO> friend Vector<NO> operator+(const Vector<NO> &a,
																								const Vector<NO> &b);
	template<uint NO> friend Vector<NO> operator-(const Vector<NO> &a,
																								const Vector<NO> &b);
	//@}
protected:
	/**
	 * Sets coordinates to zero
	 * @param start starting index to start the reset ( indexes between 0 and
	 * start-1 will not be reseted).
	 */
	void setZero(uint start = 0);

	double _[N];
};


#endif // libamarsi_VECTOR3_H_

#include "Vector-inl.h"
