/*
 * Copyright (c) 2009-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file CheetahMessages.h
 *
 * \date 16 oct. 2009
 * \author Alexandre Tuleu
 */

#ifndef _CHEETAH_MESSAGE_H_
#define _CHEETAH_MESSAGE_H_


#include <vector>

#define CUSTOM_COMMUNICATION_PROTOCOL



//#define ID_SUPERVISOR 0
#define ID_ROBOT 1
#define ID_PHYSICS 2

#include <libamarsi-quad/Message/message.hpp>
#include <libamarsi-quad/basic-math/math.hpp>

#include "global.h"


/**
 * The custom Message s of our communication protocol.
 *
 * \ingroup internal
 */
enum CustomMessageType {
//  NEW_GEOMETRY_DEFINITION = 30,
//  POSITION,
//  ORIENTATION,
//  TOGGLE_PLAY_PAUSE,
//  SIMULATION_QUIT,
  SET_MOTION_RESTRICTOR //!< SET_DEAMBULATOR Set the Deambulator states of the Cheetah
    = static_cast<unsigned int>(NB_REGISTERED_MESSAGE),
  PHYSICS_FEEDBACK,//!< PHYSICS_FEEDBACK The information returned by the physics
                  //!  plug-in at each simulation step. All are in one single
                  //!  message due to the EmitterReceiverPhysics limitation.
  SET_PHYSICS_SIMULATION_PARAMETER,
  KNEE_ORDER,
  PERTURBATION,
  DISPLAY_HINT,
  RESET_SIMULATION
};


template <uint N> Serializer & operator<<(Serializer &s, const Vector<N> & vec){
  for(uint i=0;i<N;++i)
    s<<vec[i];
  return s;
}
template <uint N> Serializer & operator>>(Serializer &s, Vector<N> & vec){
  for(uint i=0;i<N;++i)
    s>>vec[i];
  return s;
}

extern void initAllAmarsiMessages();

/**
 * Sends a change of Deambulator State.
 *
 * \ingroup internal
 */


class SetMotionRestrictor: public InternalMessage {
  friend void initAllAmarsiMessages();
  IS_A_MESSAGE(SetMotionRestrictor,SET_MOTION_RESTRICTOR)
public:
  /// value for a FREE state
  const static int FREE;
  /// value for a motion restricted to the sagital plane, with a free pitch angle.
  const static int SAGITAL_FREE_PITCH;
  /**
   * value for a motion restricted to the sagital plane, with no pitch rotation
   * permitted.
   */
  const static int SAGITAL_FIXED_PITCH;
protected:
  int t_;

public:
  virtual ~SetMotionRestrictor() {
  }
  /**
   * Sets the type of Deambulator state to enable.
   * @param t the type of Deambulator to enable
   */
  void setType(int t) {
    t_ = t;
  }
  /**
   * Returns The type of Deambulator to enable.
   * @return the type of Deambulator to enable.
   */
  int getType() const {
    return t_;
  }

  virtual Serializer & serialize(Serializer &s) {
    return s << t_;
  }

  virtual Serializer & unserialize(Serializer &s) {
    return s >> t_;
  }

};

/**
 * The Feedback returned at each simulation step by the physics plug-in.
 * It is returned in one Message due to the EmitterReceiverPhysics limitation.
 *
 * It returns :
 *  - the GRFs for each leg.
 *  - if the trunk is in contact or not
 *
 *  \ingroup internal
 */

class PhysicsFeedback : public InternalMessage{
  friend void initAllAmarsiMessages();
  IS_A_MESSAGE(PhysicsFeedback,PHYSICS_FEEDBACK)


protected :
  /**
   * \c true if the trunk is in contact with the ground.
   */
  bool d_hasABadContact;
  /**
   * an array containing the GRFs for each leg.
   */
  std::vector<Vector3> forces_;

  void checkSize(){
    if(forces_.size()==0)
      forces_.assign(4,Vector3::ZERO);
  }
public :
  virtual ~PhysicsFeedback(){}
  void setHasABadContact(bool c){
    d_hasABadContact=c;
  }

  bool hasABadContact() const{
    return d_hasABadContact;
  }

  void setForce(const Vector3 &f, unsigned int i){
    checkSize();
    if(i<forces_.size())
      forces_[i]=f;
  }

  const Vector3& getForce(unsigned int i)const{
    if(i<forces_.size())
      return forces_[i];

    return Vector3::ZERO;
  }

  std::vector<Vector3>::const_iterator  begin(){
    checkSize();
    return forces_.begin();
  }

  std::vector<Vector3>::const_iterator  end(){
    checkSize();
    return forces_.end();
  }

  virtual Serializer & serialize(Serializer &s) {
    checkSize();
    for (std::vector<Vector3>::const_iterator it = forces_.begin(); it
        != forces_.end(); ++it) {
      s << it->cx() << it->cy() << it->cz();
    }
    return s<<d_hasABadContact;
  }

  virtual Serializer & unserialize(Serializer &s) {
    checkSize();
    for (std::vector<Vector3>::iterator it = forces_.begin(); it
        != forces_.end(); ++it) {
      s >> it->x() >> it->y() >> it->z();
    }
    return s>>d_hasABadContact;
  }

  friend std::ostream & operator<<(std::ostream & out, const PhysicsFeedback &a);
};

/**
 * \ingroup internal
 */

class SetPhysicsSimulationParameter : public InternalMessage{
  friend void initAllAmarsiMessages();
  IS_A_MESSAGE(SetPhysicsSimulationParameter, SET_PHYSICS_SIMULATION_PARAMETER)
private :
  double erp_ , cfm_, friction_;
public :
  double erp() const{
    return erp_;
  }

  void setErp(double v){
    erp_=v;
  }

  double cfm() const{
    return cfm_;
  }

  void setCfm(double v){
    cfm_=v;
  }

  double friction() const{
	  return friction_;
  }

  void setFriction(double v){
	  friction_=v;
  }

  virtual Serializer & serialize(Serializer &s){
    return s<<erp_<<cfm_<<friction_;
  }

  virtual Serializer & unserialize(Serializer &s){
    return s>>erp_>>cfm_>>friction_;
  }
};
/**
 * void
 * \ingroup internal
 */
class KneeOrder : public InternalMessage {
  friend void initAllAmarsiMessages();
IS_A_MESSAGE(KneeOrder,KNEE_ORDER)
public:
  double torque() const {
    return torque_;
  }
  amarsi::Limbs limb() const {
    return static_cast<amarsi::Limbs> (limb_);
  }
  void setTorque(double v) {
    torque_ = v;
  }
  void setLimb(amarsi::Limbs l) {
    limb_ = l;
  }
  virtual Serializer & serialize(Serializer & s) {
    return s << torque_ << limb_;
  }
  virtual Serializer & unserialize(Serializer & s) {
    return s >> torque_ >> limb_;
  }
private:
  double torque_;
  int limb_;
};

class Perturbation : public InternalMessage {
  friend void initAllAmarsiMessages();
  IS_A_MESSAGE(Perturbation,PERTURBATION)
public :

  const Vector3 & force() const{
    return force_;
  }

  void setForce(const Vector3 & force){
    force_=force;
  }

  void setNumberOfSimulationStep( int v){
    nbStep_ = v<0 ? -v :v;
  }

  unsigned int numberOfSimulationStep() const{
    return nbStep_;
  }

  Serializer & serialize(Serializer & s){
    s<<force_;
    return s<<nbStep_;
  }

  Serializer & unserialize(Serializer & s){
    s>>force_;
    return s>>nbStep_;
  }
private :
  Vector3 force_;

  int nbStep_;
};


class DisplayHint : public InternalMessage {
  friend void initAllAmarsiMessages();
  IS_A_MESSAGE(DisplayHint,DISPLAY_HINT)
public :

  void setMode(amarsi::DisplayMode mode){
    mode_=mode;
  }

  amarsi::DisplayMode mode() const{
    return static_cast<amarsi::DisplayMode>(mode_);
  }

  Serializer & serialize(Serializer & s){
    return s<<mode_;
  }
  Serializer & unserialize(Serializer & s){
    return s>>mode_;
  }

private :
  int mode_;
};


class ResetMessage : public InternalMessage {
	friend void initAllAmarsiMessages();
	IS_A_MESSAGE(ResetMessage,RESET_SIMULATION)


public :

	Serializer & serialize(Serializer & s){
		return s;
	}
	Serializer & unserialize(Serializer & s){
		return s;
	}
};

#endif //_CHEETAH_MESSAGE_H_
