/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Controller.h
 *
 *  \date May 11, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_CONTROLLER_H_
#define LIBAMARSI_QUAD_CONTROLLER_H_


#include "Clock.h"
#include <string>
#include <libamarsi-quad/common/global.h>




namespace amarsi {

//forward declaration to have clean interface
struct State;
class PositionEncoder;
class Touch3DSensor;
class Accelerometer;
class Gyro;
class Actuator;
class EventManager;
class ControllerPrimple;
class Supervisor;
class Robot;
/**
 * The base class to construct a Controller upon. this class provide basic
 * access to the robot Actuator and Sensor. To run the simulation, its based
 * on almost the same API than webots through the step function. You can see the
 * BasicController example.
 *
 * The Controllers uses the Preference mechanism to define some value. When it
 * is constructed, you should provide the arguments of your program. it will
 * parse it and look for the -p/--pref preference options, that indicate
 * a path to a preference file that could add and/or redefine configuration
 * value. (see the PreferenceReader for further details).
 *
 * All the operation that could be made on the simulation, but not by the Robot
 * itself (like reseting the simulation, access to lot of inner variable of the
 * Robot (instantaneous speed, rotation, Ground Reaction Force) are accessible
 * through the Supervisor class.)
 *
 * Ideally your controller program should only have one Controller (or any
 * derived class) instantiation. Furthermore, as the model contains some
 * compliance, the physical state must be initialized before performing any
 * simulation. This is done with Supervisor£::initSimulation() method, but
 * this latter method should only be called once the Controller is created.
 * You can look into the \ref BasicControllerExample to see the architecture
 * of any controller.
 *
 *
 * \section contInit Initialization of the simulation.
 *
 * With the introduction of the management of the robot environment, the way
 * The simulation is initialized and why and when it is resetted can be hard
 * to understand.
 *
 * Has the robot contains some compliance, it dynamics must be initialized prior
 * to be used, i.e it must be put on a steady state. whenever the environment is changed
 * this also the case. A callback is provided to the user, userInitSimulation()
 * who is called just after this dynamic initialization is performed, to let the user
 * read the starting State of the robot, with all its amarsi::Sensor /
 * amarsi::Actuator initialized.
 *
 * However if the user want directly at the beginning set a certain environment,
 * this initialization will be made twice, wasting computation time. So the
 * Controller::userinitEnvironment callback has been created. At teh end we obtain
 * the following organization :
 *
 * \image html ControllerCallback.png
 *
 *
 * \ingroup user
 *
 * \sa Supervisor Actuator Sensor \ref BasicControllerExample.
 */
  class Controller {
  public:
  friend class amarsi::ControllerPrimple;
  friend class amarsi::Supervisor;
    /**
     * Standard Constructor, taking program options as an arguments.
     * @param argc the number of agruments
     * @param argv the arguments
     */
    Controller(int& argc, char ** argv);
    virtual ~Controller();



    /**
     * Almost like the webots::Robot::step methods. Perform the simulation for
     * a given time. But it do not takes the number of millisecond to run, but a
     * multiple of the basic timestep of the simulation, avoiding these annoying
     * message each time you change the basic time step of your simulation. Just
     * call it without arguments if you want to perform the smallest available
     * simulation step.
     * @param multipleOfBasicTimestep defines the time the simulation will be
     * runned as a multiple of the basic simulation step. N.B : 0 will be
     * treated as 1.
     */
    void step(unsigned int multipleOfBasicTimestep = 1);

    /**
     * Gives access to an Actuator of the robot. Each Actuator have different
     * characteristic for the command, so please refers to the model section of
     * the main libamarsi-quad documentation for further details on each Actuator.
     * @param id the ActuatorID of the wanted Actuator
     * @return a reference to the Actuator
     */
    Actuator& actuator(ActuatorID id) const;

    /**
     * Gives access to the encoder placed on the knees.
     * \sa \ref model
     * @param l the Limb of the encoder we want to access.
     * @return a reference to PositionEncoder that represents the encoder
     */
    const PositionEncoder& kneeEncoder(Limbs l) const;

    /**
     * Gives access to the encoder placed on the knees.
     * \sa \ref model
     * @param l  the Limb of the touch sensor we want to access.
     * @return a referemce to the the Touch3DSensor that represents the sensor.
     */
    const Touch3DSensor& touchSensor(Limbs l) const;

    /**
     * Gives access to the Gyro of the robot.
     */
    const Gyro & gyro() const;
    /**
     * Give access to the Accelerometer of the robot.
     * @return
     */
    const Accelerometer & accelerometer() const;
    /**
     * Actions to perform at each initialization of the simulation. Notice that
     * the purpose s to initialize/reset some inner variables of your derived
     * Controller. Therefore this method may be called several time during one
     * Webots simulation. Do not call configuration methods of the Supervisor
     * (like registering events) in this one, as they may then be registered
     * several time. See Supervisor::initSimulation() for further details.
     */
    virtual void userInitSimulation();
    /**
     * Actions to perform at each second. You can redfine this method to perform
     * certain action at each second of the simulation.
     */
    virtual void userActionEverySecond();

    /**
     * Call back to modify the environment of the robot
     * (through the Supervisor::setFlatFloor(),
     * Supervisor::setFrontStep() methods ...), whenever the simulation
     * is resetted. If you do this in userInitSimulation(), a re-initialization
     *  of the world would be needed. So you would initialize it twice, or in an
     *  infinite loop .... This callback is provided to avoid this.
     *
     *  \note The use of the above methods is disabled in the Supervisor if you
     *  are in userInitSimulation(). A warning message will be displayed.
     *
     *  \note If you call one of the method above outside both userInitSimulation()
     *  and userInitEnvironment(), then a reset of the simulation will be performed.
     *  But during the resulting initialization, userInitEnvironment() will
     *  <b>not</b> be called. It's to permit to override thes method, even if
     *  its defined. So the priority are :
     *  - Call to a change outside userInitEnvironment()
     *  - Call to a change inside useerInitEnvironment()
     *  - Flat ground
     *  - Forbidden call of environment change inside userInitSimulation()
     *
     */
    virtual void userInitEnvironment();


    /**
     * returns a pointer to the underlying robot structure.
     */
    amarsi::Robot* robot();

  private:
    void actionsEverySecond();
    void initController(int & argc, char ** argv);
    static ControllerPrimple* pimpController_;
  };

}
#endif // LIBAMARSI_QUAD_CONTROLLER_H_

