/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Event.h
 *
 *  \date May 11, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_EVENT_H_
#define LIBAMARSI_QUAD_EVENT_H_

namespace amarsi {

  /**
   * Functors that represent an Event that will be triggered at a certains Time,
   * by the vent Manager.
   *
   * You can define some Event that will occurs during the simulation at a
   * certain time, and triggered by an EventManager.
   * For that purpose you could use the built-in Supervisor::eventManager(),
   * to register such Event.
   *
   * \ingroup internal
   *
   * \sa EventManager Supervisor \ref EventExample
   *
   */
  class Event {
  public:
    Event(double time);
    virtual ~Event();

    double time() const;

    void performAction();

  protected :
    virtual void action()=0;
  private :
    double time_;

  public :
    class Comparator {
    public :
       bool operator() ( Event*  i, Event* j){
        if(i&&j)
          return (i->time()<j->time());
        return false;/// \todo should throw exception here
      }
    };
  };
}


/**
 * \example EventExample
 * These example shows how you can use Event in your controller.
 * Here we define a new event called ResetEvent that will be triggered after 10.0
 * seconds.
 *
 * We define first a standard Event header. This Event will just define the pure virtual Event::action() method
 * \include events/ResetEvent.h
 * And now there is the implementation. Note that all Event must have a defined
 * time at construction time. Here it is directly hard-coded in the constructor.
 * ResetEvent::action() just redefin
 * \include events/ResetEvent.cpp
 *
 * Now we define a dummy controller, that will have no action, but will wait
 * 10.0 second and then be reseted by our event.
 *
 * First the header
 * \include events/ResetController.h.
 * And now the implementation.
 * \include events/ResetController.cpp
 *
 * \ingroup internal
 */


#endif // LIBAMARSI_QUAD_EVENT_H_
