/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file EventManager.cpp
 *
 *  \date May 11, 2010
 *  \author tuleu
 */

#include <iostream>

#include "EventManager.h"
#include <algorithm>
#include "Clock.h"
#include "Event.h"

namespace amarsi{
  EventManager::EventManager()
    : events_()
    , nextEvent_(events_.begin()){
  }

  EventManager::~EventManager() {
    clear();
  }

  void EventManager::clear(){
    for(SetOfEvent::iterator it=events_.begin();
        it!=events_.end();
        ++it)
      delete *it;
    events_.clear();
    nextEvent_=events_.begin();
  }

  void EventManager::reset(){
    nextEvent_=events_.begin();
  }

  void EventManager::registerEvent(Event *e){

    //we test if the new event must be triggered before current one, if so
    // sorting will push forward all members, so nextEvent will now point to
    // an element which have been already triggered, to prevent this,
    // we increment it after the sorting
    bool needToIncrement(false);
    if(*nextEvent_ && (*nextEvent_)->time() > e->time())
      needToIncrement=true;
    registerEventNoSort(e);
    sort();
    if(needToIncrement)
      ++nextEvent_;
  }

  void EventManager::registerEventNoSort(Event *e){
    //attention : we need to be aware, that the vector when grows in capacity, will
    //invalidate  our iterator.
    if(events_.size()==events_.capacity()){//will grow if we push_back
      //we save the current position, should be 0, but who knows ....
      unsigned int index=nextEvent_-events_.begin();
      events_.push_back(e);//vector will grow
      nextEvent_=events_.begin()+index;//restore the correct position
    }else//no grows, just push, ouf.
      events_.push_back(e);
  }

  Event::Comparator comp;
  void EventManager::sort(){
    std::stable_sort(events_.begin(),events_.end(),comp);
  }

  void EventManager::triggerEvents(){
    while(nextEvent_!=events_.end() &&//still an event waiting
          Clock::getTime()>=(*nextEvent_)->time()){//time of event is outdated
      (*nextEvent_)->performAction();//perform event
      ++nextEvent_;//test for next one
    }
  }

  EventManager::iterator  EventManager::begin(){
  	return events_.begin();
  }

  EventManager::iterator  EventManager::end(){
  	return events_.end();
  }

  void EventManager::erase(const EventManager::iterator & beg,
                           const EventManager::iterator & end){
  	events_.reserve(events_.size());//to be sure that the array will not move.
  	events_.erase(beg,end);
  }

}
