/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file EventManager.h
 *
 *  \date May 11, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_EVENT_MANAGER_H_
#define LIBAMARSI_QUAD_EVENT_MANAGER_H_
#include <vector>

namespace amarsi{
	class Event;

  /**
   *
   * That class holds a set of Event and is responsible to trigger them at the
   * right time. Events are registered with either registerEvent and
   * registerEventNoSort . The method triggerEvents triggers all events that
   * need to be triggered. So, in order to work, this method should be called
   * at each simulation step. However the Supervisor contains an eventManager, that is automatically
   * updated, so end-user may prefer to use directly this one, rather than
   * managing the triggering manually.
   *
   * \warning This class is intended to be used as follow : register all our
   * event before clock is started, and do not modify them after. If you do so,
   * it may works, or not, but nothing is ensured. Actually it should work, if
   * you only use registerEvent() method (but registerEventNoSort() may have
   * unpredicted result)
   *
   * \sa Supervisor::eventManager() Supervisor Event EventExample
   *
   * \ingroup internal
   *
   * \todo UNIT TESTING IS HIGHLY REQUIRED HERE : lot of problem could arise from
   *  stl implementation quirks !!!
   * \todo maybe add a lock mechanism
   * \todo redo the implementation, or the complete design, as it is not enough
   * simple : The user need to know the mechanism in order not to make big
   * mistake.
   *
   */
	class EventManager {
	public :

		///@name Constructor and Destructor.
		//@{
    /**
     * Standard constructor.
     * @return
     */
    EventManager();
    virtual ~EventManager();
    //@}

    /**
     * @name Events management
     */
    /**
     * Triggers all outdated Event according to the current simulation time.
     */
    void triggerEvents();
    /**
     * Register a new Event to be triggered at the right time. The eventManager
     * own the life cycle of the passed Event.
     * @param e the new Event of be triggered
     */
    void registerEvent(Event* e);

    /**
     * Resets the triggering of all Event. All registered Events are know count
     * as non triggered, and will be trigggered if current simulation time
     * outdate them at the next call of triggerEvents.
     */
    void reset();
    /**
     * Clears all the registered Events.
     */
    void clear();



	protected :
    typedef std::vector<Event*> SetOfEvent;
    typedef SetOfEvent::iterator iterator;

    iterator begin();
    iterator end();
    void erase(const iterator & beg, const iterator & end);

    /**
     * Register A new Event, without sorting it, for efficiency. You should only
     *  use this when :
     *   - You have a bunch of Event to store in one row
     *   - Simulation is not started
     * Please remind to call sort() afterwards (if Event are not sorted, then
     * the class won't work)
     * @param e the event to register
     */
    void registerEventNoSort(Event* e);
    void sort();
  private :

    SetOfEvent events_;
    iterator nextEvent_;
  };
}
#endif // LIBAMARSI_EVENT_MANAGER_H_
