/*
 * Copyright (c) 2009-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file State.h
 *
 * \date 18 dec. 2009
 * \author Alexandre Tuleu
 */

#ifndef STATE_H_
#define STATE_H_

#include "../basic-math/math.hpp"

#include <libamarsi-quad/common/global.h>

#include <vector>

namespace amarsi {

  class LegKinematicModel;
  /**
   * A structure that define the global state of the robot.
   * At each reset(), the starting values are replaced by current one.
   * Controller insure that these value are always up-to-date outside the run()
   * method.
   *
   * \ingroup user
   *
   */
  struct State {
    /**
     * Current position.
     */
    Vector3 position;
    /**
     * Starting position, used to compute the distance traveled.
     */
    Vector3 startingPosition;
    /**
     * Position at the last simulation step.
     */
    Vector3 previousPosition;
    /**
     * Instantaneous speed.
     */
    double instantaneousSpeed;

    /**
     * Orientation at the start of the simulation, exprimed with a Quaternion
     * of the cheetah
     */
    Quat startingOrientation;
    /**
     * Current orientation of the simulation, exprimed with a Quaternion
     * of the cheetah
     */
    Quat orientation;
    /**
     * Up direction at the start of the simulation.
     */
    Vector3 startingUpDirection;
    /**
     * Front direction at the start of the simulation.
     */
    Vector3 startingFrontDirection;
    /**
     * Side direction at the start of the simulation.
     */
    Vector3 startingSideDirection;
    /**
     * Up direction of the Robot.
     */
    Vector3 upDirection;

    /**
     * Front direction of the robot.
     */
    Vector3 frontDirection;
    /**
     * Side direction of the robot.
     */
    Vector3 sideDirection;
    /**
     * \c true when the trunk is in contact with the ground. Updated by the
     * physics plug-in
     */
    bool hasABadContact;

    /**
     * Ground Reaction Forces for each Legs.
     */
    std::vector<Vector3> GRFs;

    /**
     * Consumed energy during the simulation.
     */
    double consumedEnergy;


    std::vector<LegKinematicModel*> kinematicModels_;
    State();
    ~State(){};
    /**
     * Resets all the value :
     * - speed and energy to zero
     * - starting position to current one.
     */
    void reset();

    const LegKinematicModel & legModel(amarsi::Limbs l) const;

    void updateFields(const Vector3 & newPosition,
                      const Quat& newOrientation);

    const Vector3 upDirectionOwn_;
    const Vector3 frontDirectionOwn_;
    const Vector3 sideDirectionOwn_;

  };
}

#endif /* STATE_H_ */
