/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Supervisor.cpp
 *
 *  \date Jun 2, 2010
 *  \author tuleu
 */
#include <stdlib.h>

#include <webots/Supervisor.hpp>


#include "Controller.h"
#include "EventManager.h"
#include <libamarsi-quad/common/global.h>
#include <libamarsi-quad/controller/internals/ControllerPrimpleAmarsi.h>
#include <libamarsi-quad/utils/PreferenceReader.h>
#include <libamarsi-quad/utils/Debug.h>
#include <libamarsi-quad/actuators/Actuator.h>
#include <libamarsi-quad/controller/internals/MotionRestrictorManager.h>
#include <libamarsi-quad/common/CheetahMessages.h>
#include <libamarsi-quad/floor/FloorElements.h>
#include <libamarsi-quad/benchmarks/benchmarks.h>

#include "Supervisor.h"
#include <assert.h>

#define LOCKABLE_METHOD_WITH_VALUE(returnValue)if(d_lock()){\
		return returnValue;\
}

#define LOCKABLE_METHOD if(d_lock()){\
		debug_supervisor("Try to access a locked method, but lock "<<&d_lock<<" locked"<<std::endl);\
		return ;\
}else{\
	debug_supervisor("Acceded a a lockable method and lock "<<&d_lock<<" not locked."<<std::endl);\
}


namespace amarsi {

	bool Supervisor::needInitialization_(true);
	bool Supervisor::s_isInCreation(false);
	Supervisor::Supervisor()
	: controller_(NULL), pimpController_(NULL)
	, eventManager_(new EventManager())
	, initializationTime_(3.0)
	, d_disableElementChange(true)
	, d_isInInitialization(false)
	, d_isAnUserChange(false)
	, d_lock(){
		init();
	}

	Supervisor::~Supervisor() {
		delete eventManager_;
	}

	void Supervisor::initSimulation(){
		//if we don't need an initialization, (we are running), then we don't
		//perform one
		if(!needInitialization_)
			return;
		//Controller must be created before to call it;
		d_isInInitialization = true;
		if(!controller_){
			std::cerr<<"There is no amarsi::Controller created during a call to "
					<<"amarsi::Supervisor::initSimulation(), so I quit !"<<std::endl;
			exit(1);/// \todo should throw exception here !!!
		}
		d_disableElementChange = false;//to be sure that we can chane the environment
		//callback to modify te environment.
		if(!d_isAnUserChange){//we should only modify the environment if no other is specified by the user.
#ifdef USE_LIBAMARSI_BENCH
			benchmark::WebotsSupervisor::instance().launchAtomicTestIfexternallyDriven();
			if(!benchmark::WebotsSupervisor::instance().isExternallyDriven()){
				controller_->userInitEnvironment();
			}
#else
			controller_->userInitEnvironment();
#endif



		}
		d_disableElementChange = true;

		//run the simulation, reset the lock
		pimpController_->preRunPhysics(initializationTime_);
		//resets our events if there is some
		eventManager_->reset();

		pimpController_->state_.reset();//state is now stationnary, so we reset speeds
		//and starting position


		controller_->userInitSimulation();
		d_disableElementChange = false;

		d_isInInitialization = false;
		//we are now initialized
		needInitialization_=false;
	}

	void Supervisor::resetSimulation(){
		LOCKABLE_METHOD
		//if we need an initialialization, there is no need to reset;
		if(needInitialization_)
			return;
		needInitialization_=true;
		//If we don't disable MR and we Perform the reset position,
		//simulation will crash.
		pimpController_->disableMotionRestrictor();
		//reset robot position and inertii ...
		pimpController_->resetRobotPosition();
		//we need to init
		initSimulation();
	}

	void Supervisor::simulationQuit(){
		LOCKABLE_METHOD
		pimpController_->simulationQuit();
	}

	void Supervisor::simulationQuit(unsigned int returnCode){
		LOCKABLE_METHOD
		pimpController_->simulationQuit(returnCode);
	}



	Supervisor *Supervisor::instance_(NULL);

	void Supervisor::configureMotionRestriction(const std::string & configString){
		LOCKABLE_METHOD
		pimpController_->mrManager_->setDeambulatorStates(configString);
	}

	void Supervisor::forceMotionRestrictionToFreeAfter(double time){
		pimpController_->mrManager_->forceFreeAfter(time);
	}
	//just initialize the proper pimpcontroller according to the supported models.
	// In this distribution there is only the simplified version
	void Supervisor::initPimpcontroller(webots::Supervisor *sup){
		webots::Node *robot=sup->getFromDef("AMARSI");
		if(robot)
			pimpController_ = new ControllerPrimpleAmarsi(sup,robot);
		else{
			std::cerr<<"Did not found a node called AMARSI in your world"
					<<"file, so I quit !"<<std::endl;
			exit(1);/// \todo should be an exception throw;
		}
		Controller::pimpController_=pimpController_;
	}

	void Supervisor::init(){
		PreferenceReader::instance().read("../../prefs/global.pref");
		Debug::instance().enableFromPreferences();
		webots::Supervisor *sup=new webots::Supervisor();
		initPimpcontroller(sup);
		//beware readPreferences send messages,
		//so the PimpController should be set !!!

		webots::Node *ramp(sup->getFromDef("RAMP")), *podium(sup->getFromDef("PODIUM"));


		flatFloor_= new FlatFloor(sup->getFromDef("GROUND"));
		frontSlope_= new FrontSlope(ramp,podium);
		heightMap_=new HeightMap(sup->getFromDef("HEIGHTMAP"));
		d_frontStep = new FrontStep(sup->getFromDef("STEP_FRONT"),
		                            sup->getFromDef("STEP_BACK"));
		d_slope = new Slope(sup->getFromDef("SLOPE"));
		d_stand = new Stand(sup->getFromDef("STAND"));
		readPreferences();
	}

	void Supervisor::readPreferences(){
		debug_base(PreferenceReader::instance()<<std::endl);
		PreferenceReader &r=PreferenceReader::instance();
		if(!r.setDomain("Controller"))// no defined preferences
			return;
		std::string configDeambulator;
		r.getValue("initialization_time",initializationTime_);
		if(r.setDomain("Toe")){
			SetPhysicsSimulationParameter m(ID_ROBOT,ID_PHYSICS);
			double erp,cfm,friction;
			if(!r.getValue("soft_erp",erp))
				erp=0.040;
			if(!r.getValue("soft_cfm",cfm))
				cfm=0.001;
			if(!r.getValue("friction",friction))
				friction=2.0;
			m.setErp(erp);
			m.setCfm(cfm);
			m.setFriction(friction);
			m.send_me();
		}
	}

	void Supervisor::setSimulationController(amarsi::Controller * controller){
		if(controller_){
			std::cerr<<"Warning you are creating several instance of an "
					<<"amarsi::Controller, so the must older may contains "
					<<"outdated data."<<std::endl;
		}
		pimpController_->userController_=controller;
		controller_=controller;
	}

	EventManager & Supervisor::eventManager() const {
		return *eventManager_;
	}


	const State & Supervisor::state() const {
		return pimpController_->state_;
	}

	double Supervisor::timeStepInS() const{
		std::cerr<<"Warning his method is depreciated, use amarsi::Clock::timestepInS() instead !"
		         <<std::endl;
		return Clock::timeStepInS();
	}

	int Supervisor::timeStepInMS() const{
		std::cerr<<"Warning his method is depreciated, use amarsi::Clock::timestepInMS() instead !"
		         <<std::endl;
		return Clock::timeStepInMS();
	}

	Supervisor & Supervisor::instance(){
		if(!instance_){
			if(s_isInCreation){
				std::cerr<<"Assertion failed , recursive creation of a singleton"
				         <<std::endl;
				exit(1);
			}
			s_isInCreation = true;
			instance_=new Supervisor();

#ifdef USE_LIBAMARSI_BENCH
			benchmark::WebotsSupervisorQuad::initialize();
#endif
			s_isInCreation = false;
		}
		return *instance_;
	}

	void Supervisor::setFlatFloor(){
		LOCKABLE_METHOD
		setFloorElement(flatFloor_);
	}

	void Supervisor::setRandomHeightMap(){
		LOCKABLE_METHOD
		double seed=((double)rand())/((double) RAND_MAX)*3.0 - 1.5;
		heightMap_->setSeed(seed);
		setFloorElement(heightMap_);
	}

	void Supervisor::setFrontSlope(double slope){
		LOCKABLE_METHOD
		if(slope<-0.5)
			slope=-0.5;
		else if(slope>0.5)
			slope=0.5;
		frontSlope_->setSlope(slope);
		setFloorElement(frontSlope_);
	}
	void Supervisor::setSlope(bool inFront, double slope){
		LOCKABLE_METHOD
		d_slope->setOrientation(inFront?Slope::FRONT:Slope::SIDE);
		d_slope->setSlope(slope);
		setFloorElement(d_slope);
	}
	void Supervisor::setStand(){
		LOCKABLE_METHOD
		setFloorElement(d_stand);
	}
	void Supervisor::setFrontStep(double height, double distance){
		LOCKABLE_METHOD
		d_frontStep->setDistance(distance);
		d_frontStep->setHeight(height);
		setFloorElement(d_frontStep);
	}

	void Supervisor::setFloorElement(FloorElement *e){
		LOCKABLE_METHOD
		if(d_disableElementChange){
			std::cerr<<"You try to change the simulation environment inside "
					<<"Controller::userInitializatrion(). This is forbidden. If you "
					<<"want to do this in the initialization part of your controller,"
					<<" please use the Controller::userInitEnvironment() callback."
					<<std::endl;
			return;
		}
		if(e)
			e->activate();
		if(d_isInInitialization)//we are on initSimulation(), so we skip
			return;
		//we are not in initSimulation(), so this is an user change
		d_isAnUserChange = true;
		if(needInitialization_){
			initSimulation();
		}else{
			resetSimulation();
		}
		d_isAnUserChange = false;
	}

	void Supervisor::overwriteWithNewPreferencesFile(const std::string & fileName){
		std::ifstream file(fileName.c_str());
		if(file.good())
			overwriteWithNewPreferences(file);
		/// \todo should throw exception here.
	}

	void Supervisor::overwriteWithNewPreferences(std::istream & stream){
		PreferenceReader::instance().read(stream);
		//Now as we changed the pref, we need to read it again :
		readPreferences();

		//since lot of preferences are read by the PimpController and subclasses,
		//we need to recreate them all, and reset the simulation (if needed)
		ControllerPrimple *tmp(pimpController_);
		initPimpcontroller(pimpController_->supervisor_);
		//don't forget to update the link !!!
		pimpController_->userController_=controller_;
		delete tmp;
		//reset will be ignored if we are not initialized.
		resetSimulation();
	}


	void Supervisor::addPerturbationInSagitalPlane(double force,
	                                               double duration,
	                                               double angle){


		//compute the side direction in the world coordinate
		Vector3 sideDirection(state().sideDirection);
		sideDirection.y()=0.0;

		if(sideDirection.normalize()<0.000000001)//error we are too small ...
			sideDirection=state().upDirection; //so we are in fact searching the
		//upDirection

		Quat rotation = Quat(sideDirection,angle);


		Vector3 forceVec=rotation.rotate(Vector3(force,0.0,0.0));

		//now apply everything

		addPerturbation(forceVec,duration);

	}
	void Supervisor::addPerturbation(const Vector3 & force, double duration){
		Perturbation m(ID_ROBOT,ID_PHYSICS);
		m.setForce(force);
		m.setNumberOfSimulationStep(duration/Clock::timeStepInS());
		m.send_me();
	}

	Lock::Key Supervisor::lock(){
		return d_lock.getAKey();
	}

	void Supervisor::setDisplayMode(int mode){
		DisplayHint m(ID_ROBOT,ID_PHYSICS);
		m.setMode(static_cast<amarsi::DisplayMode>(mode));
		m.send_me();
	}

	void Supervisor::setTrackingOn(bool on){
		pimpController_->setTrackingOn(on);
	}

}
