/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Supervisor.h
 *
 *  \date Jun 2, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_SUPERVISOR_H_
#define LIBAMARSI_QUAD_SUPERVISOR_H_

#ifndef NULL
  #define NULL 0
#endif

#include <iostream>
#include <cstdlib>
#include <libamarsi-quad/controller/internals/ControllerPrimple.h>
#include <libamarsi-quad/utils/Lock.h>
#include "Controller.h"

namespace amarsi {

  class EventManager;
  class Controller;
  class FloorElement;
  class FlatFloor;
  class FrontSlope;
  class HeightMap;
  class FrontStep;
  class Slope;
  class Stand;
  /**
   * A Singleton that performs all sort of operation on the robot/simulation, that
   * could not be performed on the robot alone. This include :
   *   - reset of the simulation (without restarting the webots process)
   *   - Access to a more global state of the robot (speed, orientation, true
   *   computed GRFs)
   *
   * As the robot has compliance, a stationary state of the whole dynamic of the
   * robot must be reached. For this purpose, before each repetition of the
   * simulation (i.e. at each begining of the webots simulation and after each
   * call to resetSimulation) the simulation is computed for a certain amount
   * of time with all orders setted to zero. This amount of time is defined
   * by the "initialization_time" value of the "Controller" preference domain.
   *
   * This initialization must be manually called by the user, once the Controller
   * is created, and before using anything in its interface (either Actuator,
   * or Sensor). see the \ref BasicControllerexample to see how this could be
   * implemented.
   *
   * \ingroup user
   *
   */

  class Supervisor {
    friend void Controller::initController(int & argc, char ** argv);
  public :

    /**
     * Gives a reference to the Supervisor.
     * @return a reference to the Supervisor.
     */
    static Supervisor& instance();

    static void checkIfPhysicsIsInitialized();

    /*void setController(Controller *controller,
                       PimpController *pimpController);*/


    virtual ~Supervisor();

    /**
     * The current State of the robot.
     * @return a reference to the State of the robot.
     */
    const State& state() const;
    /**
     * A reference to the EventManager handled by the Supervisor. As it is always
     * updated by the Supervisor internally, end user should prefer use this one
     * rather than maintains their own.
     * @return
     */
    EventManager& eventManager() const;

    /**
     * Initialize the physics of the simulation. As the Robot has
     * compliance, a stationary states of the physical state of the
     * robot must be be reached.  The call of this method is mandatory
     * in controller program before to run any amarsi::Controller
     * method. This call must also be made after the creation of the
     * amarsi::Controller. see the \ref BasicControllerExample for
     * further details.
     *
     * This method will run the simulation for a given time, in order
     * to reach numerically this in webots. This time is setted by the
     * \c initialization_time variable of \c Controller Preference
     * domain.
     *
     *
     *
     */
    void initSimulation();
    /**
     * Resets the simulation, by re-positionning the robot at its
     * starting point, re-initializing the physics of the simulation,
     * the EventManager states ...  A call of this method is really
     * different of a call of the webots equivalent, as the program is
     * not killed.
     */
    void resetSimulation();
    /**
     * Quits the simulation and webots. the return code will be 1.
     */
    void simulationQuit();
    /**
     * Quits the simulation and webots, like
     * webots::Supervisor::simulationQuit will do, but you can specify
     * a different return code. Please notice that if you are using a
     * webots version older than 6.3.1, then the return code will will
     * still be 1, whatever you apss as an argument. (This possibility
     * only appears since webots 6.3.1)
     */
    void simulationQuit(unsigned int returnCode);

    /**
     * Configures the MotionRestrictor with a new configuration
     * string, see the MotionManagerRestrictor documentation for the
     * correct syntax.
     * @param configString the configuration string of the MotionRestrictor.
     */
    void configureMotionRestriction(const std::string & configString);

    void forceMotionRestrictionToFreeAfter(double time);

    /**
     * Returns the basic time step of the simulation in seconds.
     * \warning this method is depreciated, use
     * amarsi::Clock::timeStepInS() instead
     * @return the time step in seconds.
     */
     double timeStepInS() const;
    /**
     * Returns the basic time step of the simulation in milliseconds.
     * \warning this method is depreciated, use amarsi::Clock::timeStepInS() instead
     * @return the time step in milliseconds
     */
     int timeStepInMS() const;

     /**
      * @name World environment controls
      * \details These methods permit the manipulation of the
      *  environment of the robot, like putting him on a slope, add a
      *  step forward him, put it on a height map.
      *
      *  \warning : using anyone of these method will re-initialize
      *  like a call to resetSimulation(). The only case this will not
      *  happen is when they are called inside
      *  Controller::userInitEnvironment(). You should also
      *  <b>never</b> call one of these methods inside
      *  Controller::userInitSimulation()
      *
      * \ingroup benchmarkTool
      */
     //@{
     /**
      * Sets the robot on a flat floor.
      *
      * \ingroup benchmarkTool
      */
     void setFlatFloor();
     /**
      * Sets on a ground with a slope in front of him. A Contrario of
      * setSlope(), the robot start on a flat gound.
      * @param slope the desired angle of the slope, in radians.
      *
      * \ingroup benchmarkTool
      */
     void setFrontSlope(double slope);
     /**
      * Sets the robot on non flat, irregular ground (a mesh). The starting
      * position on this mesh is randomly choosen.
      *
      * \ingroup benchmarkTool
      */
     void setRandomHeightMap();
     /**
      * Put the robot on a Slope. The robot start on this slope.
      * @param inFront if \c true, the slope is facing the robot, if \c false ,
      * the slope is on its side.
      * @param slope the angle of the slope.
      *
      * \ingroup benchmarkTool
      */
     void setSlope(bool inFront, double slope);
     /**
      * Puts the robot on a stand. Its feets does'nt touch the ground anymore.
      * Useful for debugging the generated trajectory of legs.
      *
      * \ingroup benchmarkTool
      */
     void setStand();
     /**
      * create a step in front of the robot a a given distance
      * @param height height of the step in meter, could also be negative
      * @param distance the distance from the projetion of the geometric center
      * of the robot to the step.
      *
      * \ingroup benchmarkTool
      */
     void setFrontStep(double height, double distance);
     //@}


     /**
      * @name Perturbation methods
      * \details These method are used to add smooth perturbation to the robot
      * on its (approximated) COG .
      *
      *
      */
     //@{
     /**
      * Perturbate the robot with the given force for the given duration.
      * @param force the force to apply
      * @param duration the duration of the perturbation.
      *
      * \ingroup benchmarkTool
      */
     void addPerturbation(const Vector3 & force, double duration);

     /**
      * Perturbate the robot by pushing it in its sagital plane.
      * @param force the amplitude of the perturbation
      * @param duration the duration of the perturbation
      * @param angle the angle of the perturbation. An angle of zero means a
      * forward push. an angle of pi/2 mean a upward push.
      *
      * \ingroup benchmarkTool
      */

     void addPerturbationInSagitalPlane(double force,
                                        double duration,
                                        double angle);
     //@}
     void overwriteWithNewPreferencesFile(const std::string & fileName);

     void overwriteWithNewPreferences(std::istream & stream);

     /**
      * Add a lock: lot of function will be locked and not gonna be used until
      * the reference is set to true. Other people trying to access the lock,
      * will get a false reference.
      * @return
      */
     Lock::Key  lock();


     /**
      * Select the a '|' combination of DisplayMode for the rendering of the
      * simulation.
      *
      * @param mode the mode to set ( bitwise '|' addition of DisplayMode).
      */
     void setDisplayMode(int mode);


     /**
      * Sets the the visual tracking of the robot on.
      *
      * @param on the tracking is setted if on is \c true, disabled otherwise.
      */
     void setTrackingOn( bool on=true);


  private :
    Supervisor();
    void setFloorElement(FloorElement *element);

    /**
     * \internal Read the values of the PreferenceReader.
     */
    void readPreferences();
    /**
     * \internal link the PimpController and the user controller together.
     * @param controller the controller to link with.
     */
    void setSimulationController(Controller* controller);
    void init();
    void initPimpcontroller(webots::Supervisor * sup);
    static Supervisor *instance_;
    static bool needInitialization_;
    static bool s_isInCreation;
    Controller *controller_;
    ControllerPrimple *pimpController_;
    EventManager *eventManager_;


    double initializationTime_;

    FlatFloor* flatFloor_;
    FrontSlope* frontSlope_;
    HeightMap* heightMap_;
    FrontStep* d_frontStep;
    Slope * d_slope;
    Stand * d_stand;
    bool d_disableElementChange;
    bool d_isInInitialization;
    bool d_isAnUserChange;

    Lock d_lock;

  };

  inline void Supervisor::checkIfPhysicsIsInitialized(){
    if(needInitialization_){
      std::cerr<<"It seem's that the physics is not correctly Initialized."
               <<"Did you call Supervisor::initSimulation() between the "
               <<"creation of your amarsi::Controller and any use of this class"
               <<"?"<<std::endl;
      exit(1);/// \todo should throw exception here !!!!!
    }

  }

}

#endif // libamarsi-quad_SUPERVISOR_H_
