/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file PimpController.cpp
 *
 *  \date May 11, 2010
 *  \author tuleu
 */

#include <iostream>

#include <webots/Supervisor.hpp>

#include <libamarsi-quad/libamarsi-quad-config.h>

#include <libamarsi-quad/controller/Controller.h>
#include <libamarsi-quad/common/CheetahMessages.h>
#include <libamarsi-quad/sensors/Sensors.h>
#include <libamarsi-quad/model/model.h>
#include <libamarsi-quad/benchmarks/benchmarks.h>
#include <libamarsi-quad/common/WebotsExplorer.h>
#include <libamarsi-quad/controller/Robot.h>
#include "MotionRestrictorManager.h"

#include "ControllerPrimple.h"

namespace amarsi {

	ControllerPrimple::ControllerPrimple(webots::Supervisor * supervisor)
	: userController_(NULL)
	, supervisor_(supervisor)
    , robot_(new amarsi::Robot(supervisor))
	, seconds_(0)
	, enableProfiling_(false)
	, d_cameraView(0)
	, d_trackingEnabled(false){
		mrManager_= new MotionRestrictorManager();
	}

	ControllerPrimple::~ControllerPrimple() {
		for(SetOfActuator::iterator act=actuators_.begin();
				act!=actuators_.end();
				++act)
			delete *act;

		for(std::vector<PositionEncoder*>::iterator enc=kneeEncoders_.begin();
				enc!=kneeEncoders_.end();
				++enc)
			delete *enc;

		for(std::vector<Touch3DSensor*>::iterator touch=touchSensors_.begin();
				touch!=touchSensors_.end();
				++touch)
			delete *touch;
		delete gyro_;
		delete accelerometer_;
		delete mrManager_;
	}


	int ControllerPrimple::step(int ms){

		//legMonitor_->updateModels();//should ever be launched before
		//the update of the actuators !!!! Value in KneeActuators depend on it
		Actuator::sendAllCommands();
		if(enableProfiling_){
			end_ = clock();
			std::cerr<<"profiling : "<<(end_-start_)<<std::endl;
		}
		supervisor_->step(ms);


		if(enableProfiling_)
			start_=clock();
		Actuator::updateAllStates();
		checkForNewMessages();//check the result of the physics plug-in
		Sensor::updateAllSensors();
		LegKinematicModel::updateAllModel();
		updateGlobals();

		//update time and trigger events
		clockTime()+=(ms)/1000.0;
		if(clockTime()>=seconds_+1.0){
			++seconds_;
			userController_->actionsEverySecond();
		}

		mrManager_->triggerEvents();

#ifdef USE_LIBAMARSI_BENCH
		benchmark::WebotsSupervisor::instance().checkCurrentBenchmarkAndAct();
#endif

		if(d_trackingEnabled){
			//check if the user have moved the camera
			Vector3 camPos(d_cameraView->getSFVec3f());
			if(!(camPos  == d_cameraPosition)){//user has moved the camera
				d_diffView = camPos - state_.previousPosition;
				d_heightToConserve = camPos.y();
			}
			d_cameraPosition = state_.position + d_diffView;
			d_cameraPosition.y()=d_heightToConserve;
			d_cameraView->setSFVec3f(d_cameraPosition.toArray());
		}

		return 0;
	}



	//  void PimpController::run(unsigned int multipleOfMilliseconds){
	//    step(multipleOfMilliseconds*timeStep_);
	//    checkForNewMessages();// first thing, because we want the info of the sensors
	//                          // that come from the physics plugin
	//    mrManager_->triggerEvents();
	//    updateGlobals();
	//  }

	void ControllerPrimple::checkForNewMessages() {
		while (Message::waiting_messages()) {
			Message m(Message::receive_message());
			if (m.is_valid() && m.am_I_receiver(ID_ROBOT)) {
				handleNewMessage(m);
			}
		}
	}


	void ControllerPrimple::handleNewMessage(Message &m) {
		switch (m.type()) {
		case PHYSICS_FEEDBACK: {
			PhysicsFeedback mm = PhysicsFeedback::cast(m);
			state_.hasABadContact = mm.hasABadContact();
			state_.GRFs.clear();//first we clean before iserting
			state_.GRFs.insert(state_.GRFs.begin(), mm.begin(), mm.end());
			//cout<<mm<<std::endl;
		}
		default:
			break;
		}
	}

	void ControllerPrimple::updateGlobals() {
		const double *pos = position_->getSFVec3f();
		//file_<<clockTime()<<" , "<<state_.position<<" , "<<state_.instantaneousSpeed<<endl;

		const double * orientation = orientation_->getSFRotation();
		state_.updateFields(Vector3(pos),
		                    Quat(Vector3(orientation[0],
		                                 orientation[1],
		                    	           orientation[2]),
		                    	   orientation[3]));
		//update the energy consumed during the step.
		for (std::vector<Actuator*>::const_iterator it = actuators_.begin();
		     it != actuators_.end();
		     ++it) {
			state_.consumedEnergy += (*it)->power() * Clock::timeStepInS();
		}
	}

	void ControllerPrimple::init(webots::Node *robotNode){
		this->setTimeStep((unsigned int) supervisor_->getBasicTimeStep());
		//enable the Message library
		initMessages();

		initActuators();
		initSensors();
		initModels();
		position_=robotNode->getField("translation");
		orientation_=robotNode->getField("rotation");

		initialPosition_=Vector3(position_->getSFVec3f());
		initialOrientation_=Vector<4>(orientation_->getSFRotation());

	}



	void ControllerPrimple::preRunPhysics(double secondes){

		for(int id=LEFT_FORE_KNEE;id<=RIGHT_HIND_KNEE;++id)
			actuators_[id]->setOrder(3.0);
		double t=0.0;
		unsigned int once(0);
		while(t<secondes){
			Actuator::sendAllCommands();
			supervisor_->step(Clock::timeStepInMS());
			Actuator::updateAllStates();
			if(once<4)
				once++;
			else if (once == 4){
				resetAllOrders();
				once++;
			}
			t+=Clock::timeStepInS();

		}

		checkForNewMessages();//check the result of the physics plug-in
		Sensor::updateAllSensors();
		LegKinematicModel::updateAllModel();
		updateGlobals();
		//update time and trigger events

		//don't forget to reset the time
		clockTime()=0.0;
		seconds_=0.0;
		//don't forget to reset the MotionRestrictorManager
		mrManager_->reset();
	}


	void ControllerPrimple::resetRobotPosition() const{

		position_->setSFVec3f(initialPosition_.toArray());

		orientation_->setSFRotation(initialOrientation_.toArray());
		ResetMessage m(ID_ROBOT,ID_PHYSICS);
		m.send_me();
		Actuator::resetAllActuators();
		resetAllOrders();
		supervisor_->simulationPhysicsReset();
	}


	void ControllerPrimple::disableMotionRestrictor(){
		SetMotionRestrictor m(ID_ROBOT,ID_PHYSICS);
		m.setType(SetMotionRestrictor::FREE);
		m.send_me();
		Actuator::sendAllCommands();
		supervisor_->step(Clock::timeStepInMS());
		Actuator::updateAllStates();
	}

	bool ControllerPrimple::messagesEnabled_(false);

	void ControllerPrimple::initMessages(){
		if(messagesEnabled_)
			return;


		initAllAmarsiMessages();

		Message::set_emitter_receiver(EmitterReceiverRobot::create(supervisor_));
		robot_->registerForbiddenName("emitter");
		robot_->registerForbiddenName("receiver");
		Message::receive_enable(Clock::timeStepInMS());

		messagesEnabled_=true;
	}


	void ControllerPrimple::setTrackingOn(bool on){
		if(!d_cameraView){
			getCameraField();
		}
		if( !(on ^ d_trackingEnabled) ){
			return;
		}

		d_trackingEnabled = true;

		if(!d_trackingEnabled)
			return;
		Vector3 camPosition(d_cameraView->getSFVec3f());
		d_diffView = camPosition - state_.position;
		d_heightToConserve = camPosition.y();

	}

	void ControllerPrimple::getCameraField(){
		webots::Node * root = supervisor_->getRoot();
		std::vector<webots::Node*> childsOfRoot;
		WebotsExplorer::getChildren(root,childsOfRoot);
		for(std::vector<webots::Node*>::iterator child = childsOfRoot.begin();
				child != childsOfRoot.end();
				++child){
			if(AMARSI_NODE_GET_TYPE_NAME((*child)) != "Viewpoint" )
				continue;

			WebotsExplorer::getField(*child,"position",&d_cameraView);

			break;
		}
	}

  amarsi::Robot* ControllerPrimple::robot(){
    return robot_;
  }

}


