/*
 * Copyright (c) 2009-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file MotionRestrictorManager.h
 *
 * \date 18 dec. 2009
 * \author Alexandre Tuleu
 */

#ifndef LIBAMARSI_MOTION_RESTRICTOR_MANAGER_H_
#define LIBAMARSI_MOTION_RESTRICTOR_MANAGER_H_

#include <string>

#include <libamarsi-quad/controller/Event.h>
#include <libamarsi-quad/controller/EventManager.h>

namespace amarsi {
  /**
   * Manages the Deambulator state of the Cheetah.
   *
   * A Deambulator can restrict the motion of the Cheetah to the sagital
   * plane, with or without a fixed pitch rotation. This Deambulator is implemented
   * in the physics plug-in, and can be activated/desactivated through
   * SetDeambulator messages.
   *
   * Then there is three states that are described in the SetDeambulator,
   * Deambulator and setDeambulatorStates() documentation.
   *
   * The DeambulatorController is here to manage these states, through a list of
   * event that are triggered at a given time. these events are defined with
   * setDeambulatorStates().
   *
   * \ingroup internal
   */
    class MotionRestrictorManager : public EventManager{
    private :

      /**
       * A single event
       */
      class Event : public amarsi::Event{
      public :
        Event(double time,int state);
        virtual ~Event();
      protected :
        virtual void action();
      private :
        int state_;
      };
    protected:
    public:
      /**
       * Standard constructor.
       */
      MotionRestrictorManager(){}
      /**
       * Standard destructor.
       */
      ~MotionRestrictorManager(){}
      /**
       * Sets the events of the controller.
       * You specify a list of times and state of the deambulator.
       * The different available state are :
       *   - "FREE" : the robot is free of movement.
       *   - "SAGITAL_FIXED_PITCH" : the robot is constrained only to translation
       *      only in the sagital frame.
       *   - "SAGITAL_FREE_PITCH" : the robot is constrained to translation and
       *      rotation in the sagital plane.
       *
       *
       * \param configurationString a string of form
       *        "XX1:YYY1|XX2:YYY2|XX3:YYY3" where :
       *    - XXi represent time of event i, in seconds.
       *    - YYYi is the state of event i.
       *
       * \todo REALLY IMPORTANT : the method should throw exception, or at least
       *       a warning when the configuration string is not valid !!!
       *
       * \note the event are not intended to be specified in causal order,
       *       the list of Event is sorted in this method.
       * \return \c true if the parse of the config string was succesful,
       *         \c false otherwise.
       */
      bool setDeambulatorStates(const std::string &configurationString);

      void forceFreeAfter(double second);
    };

}

#endif // LIBAMARSI_MOTION_RESTRICTOR_MANAGER_H_
