/*
 * Copyright (c) 2009-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Filter.h
 *
 * \date 25 nov. 2009
 * \author Alexandre Tuleu
 */

#ifndef FILTER_H_
#define FILTER_H_
/**
 * A collection of tools to perform one dimension signal processing.
 *
 *  The module consist of Filter, that perform a one dimensional causal signal
 *  processing operation. Then the signal processing could be performed like this :
\code
LowPass lp(0.5);
Threshold th(0.1);
double t(0),in,out;
for(int i=0;i<100;++i){
  t+=0.001;
  in=std::sin(t*2*PI);
  out=in>>lp>>th;
  cout<<"Input : "<<in<<" Output : "<<out<<endl;;
}
\endcode
 *  Notice than the operator use reference rather than const reference. It is because
 *  some Filter can change their inner state while performing the output, like
 *  hysteresis, derivative .... Then you should notice than performing ietratively
 *  operation of Filtering, you should not destroy the Filter Object until the end
 *  of all the computation. Indeed, Filter often need to keep track of the previous value.
 *
 *  For exemple the Following code will give an uncorrect Result :
 *
\code
double t(0),in,out;
for(int i=0;i<100;++i){
  t+=0.001;
  in=std::sin(t*2*PI);
  out=in>>LowPass(0.5)>>Threshold(0.1);
  cout<<"Input : "<<in<<" Output : "<<out<<endl;;
}
\endcode
 *
 *  as the LowPass filter is reinitialized at each iteration, and then is not able
 *  to keep track of the preceding value. At each re construction it is reset to
 * zero.
 * \ingroup internal
 */


namespace Filtering {
  class Filter;
  /**
   * Base class for Filter. Derivated class implements different types of Filters.
   */
  class Filter {
  public:
    virtual ~Filter();

    friend double  operator>>( double  input, Filter & f);
    /**
     * Applies the filter to the input as a reference. Input is altered by the
     * method
     * @param input the input to be filtered
     * @return a reference to input.
     */
     virtual double & output(double & input)=0;
    /**
     * returns the last value computed
     *
     * @return last computed value
     */
    double value() const;

  protected :
    /**
     * Standard Constructor.
     * @return a new Filter.
     */
    Filter();

    double lastValue_;

  };
}

#endif /* FILTER_H_ */
