/*
 * Copyright (c) 2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file LegKinematicModel.h
 *
 *  \date Jul 15, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_LEGKINEMATICMODEL_H_
#define LIBAMARSI_QUAD_LEGKINEMATICMODEL_H_

#include <set>

#include <libamarsi-quad/basic-math/math.hpp>

namespace amarsi {
  /**
   * A kinematic model for the leg. It is updated at each timestep. Then
   * you can access end effector position of leg at each time step through
   * positionInHipFrame(), positionInBodyFrameAtHipJoint(), positionInBodyFrame() .
   *
   * You can also compute the forward kinematic model for any position using
   * forwardKinematic() .
   *
   * You can also perform inverseKinematic using inverseKinematic() .
   *
   * The definition for the different frame used are depicted in the following
   * figure :
   *
   * \image html KinematicModelReferential.png "Referential definitaion for the provided Kinematic Model"
   *
   * \section forwardKinematic Forward Kinematic computation
   *
   * For computing the forward kinematic model, we need to settle what angle do we want
   * and what are the angle we get from webots.
   *
   * \image html LegWebotsAngles.png "Angles getted from webots"
   *
   * in the picture above we have :
   *  - \f$ \Theta_i\f$ represent the webots offset of joint i (the angle when
   *    the joint position is 0)
   *  - \f$ \theta_i \f$ the position of the webots joint (same than
   *    webots::Servo::getPosition())
   *
   * \image html LegConventionAngles.png "Angle convention used to compute forward model"
   *
   * From the two shematics it results the following realtions :
   * \f{eqnarray}{
\alpha_0 & = & \theta_0 + \Theta_0 \\
\alpha_1 & = & \theta_1 + \Theta_1 \\
\alpha_2 & = & \theta_2 + \Theta_2 + \frac{\pi}{2} \\
   *\f}
   *
   * And we get for the forward kinematic position of the end effector in the
   * Hip frame (1 in the graphic) :
\f[
 {^{1}p}=\left (\begin{matrix}
      1+\cos(\alpha_1) + \cos(\alpha_1 + \alpha_2) \\
      \sin(\alpha_1) + \sin(\alpha_1 + \alpha_2) \\
    \end{matrix}\right )
\f]

and thus we have the following formula for the body frame at hip joint :
\f{eqnarray}{
{^{0}p} &=& {^{0}R_{1}} {^{1}p}\\
               &=& \left (
                 \begin{matrix}
                   \cos\left (-\frac{\pi}{2} + \alpha_0 \right) & - \sin\left (-\frac{\pi}{2} + \alpha_0 \right) \\
                   \sin \left (-\frac{\pi}{2} + \alpha_0 \right) &  \cos \left (-\frac{\pi}{2} + \alpha_0 \right)
                 \end{matrix}
\right ) {^{1}p} \\
  {^{0}p} &=& \left (
                 \begin{matrix}
                   \sin\left (\alpha_0 \right) &  \cos\left (\alpha_0 \right) \\
                   - \cos \left (\alpha_0 \right) &  \sin \left (\alpha_0 \right)
                 \end{matrix}
\right ) {^{1}p}\\
\f}
   *
   *
   * \ingroup user
   */
  class LegKinematicModel {
  public:
    static void updateAllModel();

    LegKinematicModel();
    virtual ~LegKinematicModel();

    /**
     * Gets the end effector position in the frame relative to the Hip segment.
     * @return a 2 dimensionnal vector in the sagital plane.
     */
    const Vector<2> & positionInHipFrame() const;
    /**
     * Gets the end effector position in the frame relative to the main Body,
     * but translated at the position of the HipJoint.
     * @return a 2 dimensionnal vector in the sagital plane.
     */
    const Vector<2> & positionInBodyFrameAtHipJoint() const;

    /**
     * Gets the end effector position in the frame relative to the main Body.
     * @return a 3 dimmensionnal vector in the ain body frame.
     */
    const Vector3 & positionInBodyFrame() const;

    /**
     * Gets the corresponding length of the leg.
     */
    double length() const;
    /**
     * Gets the relative length of the leg.
     * @return
     */
    double relativeLength() const;
    /**
     * Gets the equivalent angle of the leg.
     * @return
     */
    double angle();

    /**
     * Computes the forward kinematic model for one given position.
     * The intended hipAngle are the one given by the Hips Actuator .
     * The intended kneeAngle are the one we got back from the PositionEncoder.
     * Note that if your provide other angle it will end up with wrong values
     * @param hipAngle
     * @param kneeAngle
     * @param positionInHipFrame
     * @param positionInBodyFrameAthipJoint
     * @param positionInBodyFrame
     */
    virtual void forwardKinematic(double hipAngle, double kneeAngle,
                                  Vector<2> & positionInHipFrame,
                                  Vector<2> & positionInBodyFrameAthipJoint,
                                  Vector3 & positionInBodyFrame) const =0;
    //    /**
    //     * Computes the inverse kinematic model for one joint. If it fails to found
    //     * a configuration, it will return false.
    //     * @param positionInBodyFrame the expected end effector position
    //     * @param hipAngle the corresponding order for the hip Actuator .
    //     * @param kneeAngle the corresponding order for the knee Actuator
    //     * @return \c true if succesfully found a corresponding configuration.
    //     */
    //    bool inverseKinematics(const Vector3 & positionInBodyFrame,
    //                           double & hipAngle,
    //                           double & kneeAngle);
    /**
     * Computes the inverse kinematic model for one joint. If it fails to found
     * a configuration, it will return false.
     *
     *
     *
     * @param positionInBodyFrame the expected end effector position
     * @param hipAngle the corresponding order for the hip Actuator .
     * @param kneeAngle the corresponding order for the knee Actuator
     * @return \c true if succesfully found a corresponding configuration.
     */
    virtual bool
        inverseKinematics(const Vector<2> & positionInBodyFrameAtHipJoint,
                          double & hipAngle, double & kneeAngle)const =0;

  protected:
    virtual void update()=0;
    Vector<2> d_endEffectorInHip;
    Vector<2> d_endeffectorInBodyHip;
    Vector3 d_endEffectorInBody;

    double d_length;
    double d_angle;
    double d_relativeLength;
    double d_maxLength;

  private:
    typedef std::set<LegKinematicModel*> SetOfModel;

    static SetOfModel s_models;

  };

  inline const Vector<2> & LegKinematicModel::positionInHipFrame() const {
    return d_endEffectorInHip;
  }
  inline const Vector<2> & LegKinematicModel::positionInBodyFrameAtHipJoint() const {
    return d_endeffectorInBodyHip;
  }
  inline const Vector3 & LegKinematicModel::positionInBodyFrame() const {
    return d_endEffectorInBody;
  }
  inline double LegKinematicModel::length() const {
    return d_length;
  }
  inline double LegKinematicModel::relativeLength() const {
    return d_relativeLength;
  }
  inline double LegKinematicModel::angle() {
    return d_angle;
  }
}
#endif // LIBAMARSI_QUAD_LEGKINEMATICMODEL_H_
