/*
 * Copyright (c) 2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file LegKinematicModel.cpp
 *
 *  \date Jul 14, 2010
 *  \author tuleu
 */

#include <iostream>
#include <cmath>

#include <webots/Node.hpp>
#include <webots/Servo.hpp>

#include <libamarsi-quad/libamarsi-quad-config.h>

#include "LegKinematicModelAmarsi.h"



namespace amarsi {

  webots::Field * LegKinematicModelAmarsi::getField(webots::Node* n,
                                              const std::string & name){
    webots::Field * ret = n->getField(name);
    if(!ret){/// \todo should throw an exception.
      std::cerr<<"Node '"<<AMARSI_NODE_GET_TYPE_NAME(n)<<"' should have a '"
               <<name<<"'  but not, I exit"<<std::endl;
      clear();
      exit(1);
    }
    return ret;
}

  void LegKinematicModelAmarsi::init(webots::Node * hipNode,
                               webots::Node * knee1Node,
                               webots::Node * knee2Node,
                               unsigned int timeStep){
    /// \todo should throw exception if servo are null ...

    d_hip->enablePosition(timeStep);
    d_knee1->enablePosition(timeStep);
    d_knee2->enablePosition(timeStep);


    webots::Field *hipTranslation = getField(hipNode,"translation");
    webots::Field *hipRotation = getField(hipNode,"rotation");
    webots::Field *knee1Rotation = getField(knee1Node,"rotation");
    webots::Field *knee1Translation = getField(knee1Node,"translation");
    webots::Field *knee2Rotation = getField(knee2Node,"rotation");

    d_segmentLength=std::abs(knee1Translation->getSFVec3f()[1]);
    d_alpha0TotalOffset=hipRotation->getSFRotation()[3];
    d_alpha1TotalOffset =knee1Rotation->getSFRotation()[3];
    d_alpha2TotalOffset =knee2Rotation->getSFRotation()[3];
    d_hipPositionInBodyFrame = Vector3(hipTranslation->getSFVec3f());

    d_hipMinPosition = getField(hipNode,"minPosition")->getSFFloat();
    d_hipMaxPosition = getField(hipNode,"maxPosition")->getSFFloat();
    d_kneeMinPosition = getField(knee2Node,"minPosition")->getSFFloat();
    d_kneeMaxPosition = getField(knee2Node,"maxPosition")->getSFFloat();


    forwardKinematic(0.0,0.0,0.0,d_endEffectorInHip,
                     d_endeffectorInBodyHip, d_endEffectorInBody);

    d_maxLength = d_endEffectorInHip.norm();
    updateEquivalent();

  }


  LegKinematicModelAmarsi::LegKinematicModelAmarsi(webots::Servo *hip,
                                      webots::Servo *knee1,
                                      webots::Servo *knee2,
                                      webots::Node *hipNode,
                                      webots::Node *knee1Node,
                                      webots::Node *knee2Node,
                                      unsigned int timeStep)
  : d_hip(hip)
  , d_knee1(knee1)
  , d_knee2(knee2){
    init(hipNode,knee1Node,knee2Node,timeStep);
  }

  void LegKinematicModelAmarsi::clear(){
  }

  LegKinematicModelAmarsi::~LegKinematicModelAmarsi() {
    clear();
  }

//  bool LegKinematicModel::inverseKinematics(const Vector3 & positionInBodyFrame,
//                                            double & hipAngle,
//                                            double & kneeAngle) {
//  }

  void LegKinematicModelAmarsi::forwardKinematic(double hipAngle,
                                           double knee1Angle,
                                           double knee2Angle,
                                           Vector<2> & positionInHipFrame,
                                           Vector<2> & positionInBodyFrameAthipJoint,
                                           Vector3 & positionInBodyFrame) const{
    double alpha0 = hipAngle + d_alpha0TotalOffset ;
    double alpha1 = knee1Angle + d_alpha1TotalOffset ;
    double alpha2 = knee2Angle + d_alpha2TotalOffset + M_PI/2.0;
    double beta = alpha2 + alpha1;

    positionInHipFrame [0] = (1 + std::cos(alpha1) + std::cos(beta)); //x_1
    positionInHipFrame [1] = (    std::sin(alpha1) + std::sin(beta)); //y_1
    // now the length
    positionInHipFrame *= d_segmentLength;

    //x_0
    positionInBodyFrameAthipJoint [0] = std::sin(alpha0) * positionInHipFrame[0]
                                        + std::cos(alpha0) * positionInHipFrame [1];
    // y_0
    positionInBodyFrameAthipJoint [1] = std::sin(alpha0) * positionInHipFrame[1]
                                        - std::cos(alpha0) * positionInHipFrame [0];
    //just add the offset.
    positionInBodyFrame = (d_hipPositionInBodyFrame + Vector3(positionInBodyFrameAthipJoint));

  }

  void LegKinematicModelAmarsi::updateEquivalent(){
    d_length = d_endEffectorInHip.norm();
    d_relativeLength = d_length / d_maxLength;
    d_angle = std::atan2(d_endeffectorInBodyHip[1],d_endeffectorInBodyHip[0]);
  }

  void LegKinematicModelAmarsi::update() {
    forwardKinematic(d_hip->getPosition(),
                     d_knee1->getPosition(),
                     d_knee2->getPosition(),
                     d_endEffectorInHip,
                     d_endeffectorInBodyHip,
                     d_endEffectorInBody);
    updateEquivalent();
  }

  bool LegKinematicModelAmarsi::inverseKinematics(
                                            const Vector<2> & positionInBodyFrameAtHipJoint,
                                            double & hipAngle,
                                            double & kneeAngle) const{
    double norm2 = positionInBodyFrameAtHipJoint.norm2();
    if (std::sqrt(norm2) > d_maxLength)
      return false;//we want to go too far away.
    norm2 /= d_segmentLength * d_segmentLength;
    double alpha1 = std::acos(-(norm2 - 5.0) / 4.0);
    kneeAngle = -alpha1 + d_alpha2TotalOffset;

    if(kneeAngle < d_kneeMinPosition || kneeAngle > d_kneeMaxPosition)
      return false;

    double targetAngle = std::atan2(positionInBodyFrameAtHipJoint[1],
        positionInBodyFrameAtHipJoint[0]);

    double angleInHip = std::atan2(-std::sin(alpha1), 2 - std::cos(alpha1));

    hipAngle = targetAngle - angleInHip - d_alpha0TotalOffset;

    if(hipAngle < d_hipMinPosition || hipAngle > d_hipMaxPosition)
      return false;

    return true;
  }

  void LegKinematicModelAmarsi::forwardKinematic(double hipAngle,
                                                 double kneeAngle,
                                                 Vector<2> & positionInHipFrame,
                                                 Vector<2> & positionInBodyFrameAthipJoint,
                                                 Vector3 & positionInBodyFrame) const{
    forwardKinematic(hipAngle,-kneeAngle,kneeAngle,
                     positionInHipFrame,positionInBodyFrameAthipJoint,
                     positionInBodyFrame);
  }

}
