/*
 * Copyright (c) 2010 EPFL (Ecole Polytechnique Fédérale de Lausanne)
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Accelerometer.h
 *
 *  \date Jun 18, 2010
 *  \author tuleu
 */

#ifndef LIBAMARSI_QUAD_ACCELEROMETER_H_
#define LIBAMARSI_QUAD_ACCELEROMETER_H_

#include "Sensor.ih"
#include <libamarsi-quad/basic-math/Vector3.h>
#include <libamarsi-quad/filtering/Discretizer.h>


namespace webots {
  class Accelerometer;
  class Field;
}




namespace amarsi {

  /**
   * An accelerometer, that determines relative displacement with a simple
   * euler integration.
   *
   * This class is all purpose class, if you want the specifications of the
   * simplified Cheetah model please refers to \ref sensors and \ref model .
   * \ingroup sensors
   * \ingroup user
   *
   */
  class Accelerometer : public amarsi::Sensor {
  public:
    ///@name Constructors and Destructor
    //@{
    Accelerometer(webots::Accelerometer * accelerometer, webots::Field * gravity,
                  double timestep);
    virtual ~Accelerometer();
    //@}

    ///@name accessors
    //{
    /**
     * acceleration during the current simulation step.
     * @return a reference to the acceleration for the simulation step.
     */
    const Vector3 & acceleration() const;
    /**
     * Integrated velocity during the current simulation step. This is reseted by
     * reset(), and each time the simulation is reseted by Supervisor::simulationReset()
     * @return a reference to the speed for the current simulation step.
     */
    const Vector3 & speed() const;
    /**
     * Integrated relative displacement for the current simulation step. This is
     * reseted by reset(), and each time the simulation is reseted by
     * Supervisor::simulationReset()
     * @return a reference to the relative position during the current
     *         simulation step.
     */
    const Vector3 & displacement() const;
    //}
    /**
     * Resets all the integrated values (speed and displacement).
     */
    void reset();
  protected :
    virtual void update();
  private :
    void init();
    webots::Accelerometer * acc_;
    double timestep_;
    Vector3 acceleration_,speed_,relativePosition_;
    Vector3 correction_;
  };


//==============================================================================
// INLINE IMPLEMENTATION
//==============================================================================

  inline const Vector3 & Accelerometer::acceleration() const {
    return acceleration_;
  }

  inline const Vector3 & Accelerometer::speed() const{
    return speed_;
  }

  inline const Vector3 & Accelerometer::displacement() const{
    return relativePosition_;
  }

  inline void Accelerometer::reset(){
    speed_=Vector3::ZERO;
    relativePosition_=Vector3::ZERO;
  }

}

std::ostream & operator<<(std::ostream & out,
                          const amarsi::Accelerometer & acc);

#endif // libamarsi-quad_ACCELEROMETER_H_
