/*
 * Copyright (c) 2009-2010 Alexandre Tuleu
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Touch3DSensor.h
 *
 * \date 18 d�c. 2009
 * \author Alexandre Tuleu
 */

#ifndef TOUCH3DSENSOR_H_
#define TOUCH3DSENSOR_H_

#include "Sensor.ih"
#include <libamarsi-quad/basic-math/math.hpp>
#include <libamarsi-quad/filtering/filtering.h>

namespace amarsi{
  class Touch3DSensor;
}
std::ostream & operator<<(std::ostream & out,
                          const amarsi::Touch3DSensor & touch);

namespace amarsi {
  /**
   * A simple Touch sensor, but not based a webots::Sensor. Takes its value from
   * the Physics plug-in. The contact() output is filtered by LowPass and
   * Threshold filters. The vector value is also discretized.
   *
   * This class is all purpose class, if you want the specifications of the
   * simplified Cheetah model please refers to \ref sensors and \ref model .
   * \ingroup sensors
   * \ingroup user
   *
   */
  class Touch3DSensor: public Sensor {
  public:
    friend std::ostream & ::operator<<(std::ostream & out,
                                       const amarsi::Touch3DSensor & touch);


    /**
     *
     * Creates a Touch3DSensor
     * @param value a reference to the value to use for the sensor. this value
     *        should be externally updated at each timestep
     * @param lpAmount ammount of LowPass
     * @param threshold Threshold to use
     * @param min minimum value to be returned by the sensor
     * @param max maximal value to be returned by the sensor
     * @param precision number of bit with which the value will be encoded
     * @return
     */
    Touch3DSensor(const Vector3& value, double lpAmount,
                  double threshold, double min, double max,
                  unsigned int precision);
    virtual ~Touch3DSensor();

    /**
     * Returns the force on the leg.
     * @return A Vector3 containing the computed force by the background.
     */
    const Vector3& force() const;
    /**
     * A filtered output that detects contact with the ground
     * @return \c true if the sensor is in contact with the ground.
     */
    bool contact() const;
 protected :
    virtual void update();
 private :
    const Vector3& value_;
    Vector3 discretizedValue_;
    bool contact_;
    Filtering::LowPass lp_;
    Filtering::Threshold th_;
    Filtering::Discretizer d_;
  };

  inline const Vector3& Touch3DSensor::force() const{
    return discretizedValue_;
  }

  inline bool Touch3DSensor::contact() const{
    return contact_;
  }
}

#endif /* TOUCH3DSENSOR_H_ */
