/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Debug.cpp
 *
 *  \date Jun 10, 2010
 *  \author Alexandre Tuleu
 */
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <cctype>
#include <algorithm>
#include "libamarsi-quad/utils/Debug.h"
#include "libamarsi-quad/utils/PreferenceReader.h"




#ifdef WIN32 // to avoid some warning under windows ....
  #define get_env_variable( return_var , name ) size_t len ;\
  _dupenv_s(& return_var , &len , name )
#else
  #define get_env_variable( return_var , name ) return_var = getenv(name)
#endif

Debug* Debug::instance_(NULL);

std::ostream Debug::Domain::noOutput_(new Debug::StreamBuf());


Debug::Domain::Domain()
 : isEnabled(false)
 , output(&noOutput_)
 , name()
 , id(0){
}

Debug::Domain::Domain(bool state_, std::ostream & stream,
                      const std::string & name_)
 : isEnabled(state_)
 , output(&stream)
 , name(name_)
 , id(0){
  std::transform(name.begin(),name.end(),name.begin(),toupper);
}



Debug::Debug(){
  //I don't know why, but gcc doesn't want to use a temporary variable here,
  //soo ....
  Domain none(false,std::cerr,"NONE");
  addDomain(none);
  Domain base(false,std::cerr,"BASE");
  addDomain(base);
  Domain controller(false,std::cerr,"CONTROLLER");
  addDomain(controller);
  Domain physics(false,std::cerr,"PHYSICS");
  addDomain(physics);
  Domain supervisor(false,std::cerr,"SUPERVISOR");
  addDomain(supervisor);

}




Debug::~Debug() {
  for(SetOfFiles::iterator file=files_.begin();file!=files_.end();++file){
    (*file)->close();
    delete *file;
  }
}

Debug::Domain & Debug::domain(Debug::Domain::ID id){
  if(id<domains_.size())
    return domains_[id];
  return domains_[0];// \todo should throw exception, or not
}

std::ostream & Debug::out(Debug::Domain::ID id){
  return instance().domain(id).isEnabled ? *(instance().domain(id).output)
                                     : Domain::noOutput_;
}


Debug::Domain::ID Debug::addDomain(Debug::Domain & domain_){
  domain_.id=domains_.size();
  domains_.push_back(domain_);
  return domains_.size()-1;
}

Debug & Debug::instance(){
  if(!instance_)
    instance_=new Debug();
  return *instance_;
}

void Debug::setOutput(Debug::Domain::ID id, std::ostream & output){
  domain(id).output=&output;
}
void Debug::setOutput(Debug::Domain::ID id, std::ofstream * output){
  files_.insert(output);//to be cleaned at the end of program
  domain(id).output=output;
}

void Debug::enable(Domain::ID id){
  domain(id).isEnabled=true;
}

void Debug::disable(Domain::ID id){
  domain(id).isEnabled=false;
}

void Debug::enableAllDomains(){
  for(SetOfDomain::iterator domainIt=domains_.begin();
      domainIt!=domains_.end();
      ++domainIt)
    domainIt->isEnabled=true;
}
void Debug::disableAllDomains(){
  for(SetOfDomain::iterator domainIt=domains_.begin();
      domainIt!=domains_.end();
      ++domainIt)
    domainIt->isEnabled=false;
}

void Debug::enableFromEnvironment() {
  char *rEnv;
  MapFile usedFile;
  for (SetOfDomain::iterator domainIt = domains_.begin();
      domainIt!= domains_.end(); ++domainIt) {
    std::ostringstream envVariable;
    envVariable<<"DEBUG_"<<domainIt->name<<std::flush;
    get_env_variable(rEnv,envVariable.str().c_str());
    if(!rEnv){
      domainIt->isEnabled=false;
      continue;
    }
    domainIt->isEnabled=true;
    std::istringstream iss(rEnv);
    unsigned int level(0);
    iss>>level;
    if(level>0){
      domainIt->output=&std::cerr;
    }else{
      MapFile::const_iterator fileIt=usedFile.find(rEnv);
      if(fileIt==usedFile.end()){//open a new file
        std::ofstream *file=new std::ofstream(rEnv);
        setOutput(domainIt->id,file);
        usedFile.insert(std::make_pair(std::string(rEnv),file));
      }else{//file already opened, don't reopen it just used the pointer to the
        //right stream
        setOutput(domainIt->id,fileIt->second);
      }
    }
  }
}


  void Debug::setADomainAccordingPreferences(Debug::Domain & inspectedDomain,
                                   MapFile & usedFile ){
    std::string outValue;
    std::string variable_name(inspectedDomain.name);
    std::transform(variable_name.begin(),variable_name.end(),
                   variable_name.begin(),tolower);
    variable_name.append("_output");
    debug_none("Searching the preferences for '"<<inspectedDomain.name
                <<"' with preference variable '"<<variable_name<<"' ... ");
    if(!PreferenceReader::instance().getValue(variable_name,outValue)){
      debug_out_nh(NONE,"not found, use default output and disable domain."<<std::endl);
      inspectedDomain.isEnabled=false;
      return;
    }
    inspectedDomain.isEnabled=true;
    if(outValue=="STD_ERR"){
      debug_out_nh(NONE,"found, use standard error."<<std::endl);
      setOutput(inspectedDomain.id,std::cerr);
      return;
    }
    if(outValue=="STD_OUT"){
      debug_out_nh(NONE,"found, use standard output."<<std::endl);
      setOutput(inspectedDomain.id,std::cout);
      return;
    }
    debug_out_nh(NONE,"found, use file '"<<outValue<<"'."<<std::endl);
    MapFile::const_iterator fi=usedFile.find(outValue);

    if(fi==usedFile.end()){//file not already used, so we add it
      std::ofstream *file= new std::ofstream(outValue.c_str());
      usedFile.insert(std::make_pair(outValue,file));
      setOutput(inspectedDomain.id,file);
    }else{//file already used
      setOutput(inspectedDomain.id,fi->second);
    }

  }


  void Debug::enableFromPreferences(){
    debug_none("Using preferences for getting debug information"<<std::endl);

    if(!PreferenceReader::instance().setDomain("Debug"))
      return;//Domain Debug not defined, just silently go out

    MapFile usedFiles;
    for(SetOfDomain::iterator dom=domains_.begin()+1;
        dom!=domains_.end();
        ++dom)
      setADomainAccordingPreferences(*dom,usedFiles);
    setADomainAccordingPreferences(domains_[0],usedFiles);

    bool disableAll,enableAll;
    if(PreferenceReader::instance().getValue("disable_all",disableAll)
        && disableAll){
      disableAllDomains();
      debug_none("All domains desactivated");
    }
    if(PreferenceReader::instance().getValue("enable_all",enableAll)
        && enableAll){
      enableAllDomains();
      debug_none("All domains activated");
    }
  }

    Debug::Domain::ID Debug::addDomain(bool state,
                                       std::ostream & output,
                                       const std::string & name){
      //add manually the temporary variable to avoid some gcc problem.
      Domain tmp(state,output,name);
      return addDomain(tmp);
    }

/**
 * \example DebugExample
 *
 */
