/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/**
 * \file Debug.h
 *
 *  \date Jun 10, 2010
 *  \author Alexandre Tuleu
 */

#ifndef LIBAMARSI_QUAD_DEBUG_H_
#define LIBAMARSI_QUAD_DEBUG_H_

#include <ostream>
#include <vector>
#include <string>
#include <map>
#include <set>


/**
 * \defgroup Debug Debug module.
 *
 *  This module contains classes and macros that are useful for organizing the
 *  debug outout of the program. Debug can be organized in "Domain", each output
 *  can specify in which file and line it had been launched. Each Domain can be
 *  activated/desacivated independently. Each Domain can be either be assigned
 *  to a standard stream (cout,cerr, ... ) or a specific files. Moreover it works
 *  well with the Preference module.
 *
 * \ingroup user
 *
 */
//@{

/**
 * \name Macro output function.
 *
 * User of this Module should use these macro for the debuging output. These macro
 * are defined under the DEBUG guard symbol. So if DEBUG preprocessing macro is
 * not defined, all of the debug output of your program will be disabled.
 */
//@{
#ifdef DEBUG
  /**
   * Defines the header for a certain Debug::Domain.
   */
  #define debug_header(domain) "["<< #domain <<"]"<<__FILE__<<":"<<__LINE__<<" : "
  /**
   * Searches for the correct stream output associated with a domain, and append
   * debug_header and the passed std::ostream to it.
   * should be used like this in your code :
\code
  debug_(BASE,"I want to add this to the Debug::Domain::BASE output."
              <<std::endl);
\endcode
   */
  #define debug_out(domain, ostream) Debug::out(Debug::Domain::domain)<<debug_header(domain)<<ostream
  /**
   * Just the same as debug_out, but do not display the debug_header
   */
  #define debug_out_nh(domain , ostream) Debug::out(Debug::Domain::domain)<<ostream
#else //DEBUG
  #define debug_header(domain)
  #define debug_out(domain, ostream )
  #define debug_out_nh(domain , ostream)
#endif //DEBUG

/**
 * Just a standard shortcut for outputting on the Debug::Domain::NONE Domain.
 */
#define debug_none(stream) debug_out(NONE, stream )
/**
 * Just a standard shortcut for outputting on the Debug::Domain::BASE Domain.
 */
#define debug_base(stream) debug_out(BASE, stream )
/**
 * Just a standard shortcut for outputting on the Debug::Domain::CONTROLLER
 * Domain.
 */
#define debug_controller(stream) debug_out(CONTROLLER , stream )
/**
 * Just a standard shortcut for outputting on the Debug::Domain::PHYSICS Domain.
 */
#define debug_physics(stream) debug_out(PHYSICS , stream )

/**
 *
 */
#define debug_supervisor(stream) debug_out(SUPERVISOR, stream)
//@}
//@}
/**
 * Main class used for debug output.
 *
 * \section Presentation
 *
 * This class work together with a number
 * of preprocessor macros debug_out(), debug_out_nh() .
 * The idea behind is to :
 *
 *  - cut all the debug output in Debug::Domain. Each Debug::Domain can be
 *  activated/desactivated independently, and associated with a std::ostream.
 *  For example some of the programmer debugging output can be
 * directed to std::cerr and other to a log file.
 *
 *  - disable all the Debugging output call easily : if the code is compiled
 *  without the DEBUG pre-processing macro defined, then all call to debug_xxx
 *  macro will be removed, and not even compiled.
 *
 * Different Debug::Domain could point to the same file, if you use the proper
 * setOutput to do it (the one with an std::ofstream* as a parameter).
 *
 * There are several predefined Debug::Domain :
 *   - "NONE"
 *   - "BASE"
 *   - "CONTROLLER"
 *   - "PHYSICS"
 *
 * By default, all outputs to std::cerr, and are disabled, at the exception of
 * NONE which is enabled by default.
 *
 * \section ManageDomain Managements of the Domain
 *
 * All the enabling/disabling and association with the output o a particular
 * domain can be configured in three way :
 *  \subsection Manually
 *
 *  Just manuipulate Debug::Domain through their
 *  Debug::Domain::ID and the enable(), disable() and setOutput() methods
 *
 *  \subsection env With Environnement variables
 *
 *  Just call enableFromEnvironment() : then all Domain will be configured this
 *  way :
 *      - they all are disabled by default.
 *      - they becomes enabled if DEBUG_XXX (XXX being the domain name) variable
 *        is set.
 *        - their output is directed to std::cerr if DEBUG_XXX is a number
 *          greater than 0.
 *        - their output is directed to a certain file, whose path is DEBUG_XXX.
 *  \subsection With the preference module.
 *
 *  Just use the enableFromPreferences() static method. Then, according the value
 *  of the following variables in "Debug" Preference PrefrenceReader::Domain :
 *      - all Debug::Domain are disabled
 *      - if the value of "xxx_output" (where xxx is the *lowercase* name of the
 *        domain) is setted, then the corresponding Debug::Domain is enabled,
 *        and is directed either to :
 *        - std::cout if value is \c STD_OUT
 *        - std::cerr if value is \c STD_ERR
 *        - a file defined by the path provided by the value of the variable.
 *      - all Debug::Domain are disabled if "disable_all" is true
 *      - all Debug::Domain are enabled (overriding "disable_all") if
 *        "enable_all" is \c true
 *
 *
 * \section AddDomain Adding new Domain
 *
 * It is also possible for the end user to add a new Debug::Domain. If the new
 * defined Debug::Domain is added before the call to any
 * enableFromPreferences/enableFromEnvironment, then it will also be handled by
 * these two methods. Please also remind to add the DEBUG pre-processing macro at
 * compilation time (by using the -DDEBUG flags under gcc for example) to allow
 * the output of the Debugging information.
 *
 * \ingroup Debug
 *
 * \sa DebugExample
 */

class Debug {
public :
  /**
   * Defines a Domain for debugging. A Domain could be enabled and disabled, and
   * associated with a particular output stream.
   */
  class Domain{
  public :
    friend class Debug;
    /**
     * ID of a Domain.
     */
    typedef unsigned int ID;
    /**
     * Predefined domains
     */
    enum Type{
      NONE=0,    //!< NONE "NONE" Domain, points to std::cerr and is <b>enabled
                 //!<       </b> by default.
      BASE,      //!< BASE "BASE" Domain, points to std::cerr and is disabled by
                 //!<      default.
      CONTROLLER,//!< CONTROLLER "CONTROLLER" Domain, points to std::cerr and is
                 //!<             disabled by default.
      PHYSICS,   //!< PHYSICS "PHYSICS" Domain, points to std::cerr and is
                 //!<          disabled by default.
      SUPERVISOR,
      USER//should always be the last ....
    };
  private :
    Domain(bool state, std::ostream & output, const std::string & name);
    Domain();
    /**
     * States if the domain isEnabled or not. If enabled, debug data will be
     * passed to output
     */
    bool isEnabled;
    /**
     * output used by the domain.
     */
    std::ostream* output;
    /**
     * name of the domain
     */
    std::string name;
    /**
     * ID of the domain
     */
    ID id;
    static std::ostream noOutput_;
  };

  /**
   * Acces to the instance of the Debug class.
   * @return
   */
  static Debug & instance();
  /**
   * Returns the correct output for a given Debug::Domain. If the Domain is
   * disabled, then it returns an std::ostream reference that point to nothing.
   * @param domain The Debug::Domain::ID of the domain
   * @return A reference to the correct stream according to the Debug::Domain
   * state.
   */
  static std::ostream & out(Domain::ID domain);

  /**
   * Adds a new Debug::Domain
   * @param state the default state of the Domain
   * @param output its default associated output
   * @param name the name of the domain. by convention, it would be converted
   * to uppercase.
   * @return the associated Debug::Domain::ID of the Debug::Domain.
   */
  Domain::ID addDomain(bool state, std::ostream & output, const std::string & name);


  ///@name Domain Manipulation Methods
  //@{
  /**
   * Sets the output of the domain to a certain stream. Notes that the
   * end-user needs to close any resources by itself. For std::fstream, it is
   * better to use  the other setOutput(Debug::Domain::ID, std::ofstream *)
   * method.
   * @param domain the Debug::Domain to modify.
   * @param stream the new stream to use.
   */
  void setOutput(Domain::ID domain , std::ostream & stream);
  /**
   * Sets the output of the domain to certain file stream. The Debug classe is
   * also responsible for the std::ofstream life cycle.
   * @param domain specifies the Debug::Domain to change.
   * @param stream pointer to the std::ofstream to use for output. File will
   * be closed by Debug when needed.
   */
  void setOutput(Domain::ID domain , std::ofstream * stream);

  /**
   * Enables all the Debug::Domain.
   */
  void enableAllDomains();
  /**
   * Disables all the Debug::Domain.
   */
  void disableAllDomains();
  /**
   * Enables all the Debug::Domain according to the DEBUG_XXX environement
   * variables (where XXX is the uppercase version of Debug::Domain::name).
   *   - If DEBUG_XXX is a number greater than 0, Domain will be activated and
   *   it will point to std::cerr
   *   - It will point to a file defined by the path DEBUG_XXX otherwise.
   */
  void enableFromEnvironment();
  /**
   * Enables all the Debug::Domain according to the values of certain variables
   * in the "Debug" preferences Domain. specific domain variables are xxx_output
   * , where xxx is the name lowercase version of Debug::Domain::name)
   *   - If "xxx_output" is setted, then the domain will be activated
   *   - If "xxx_output" is setted to either "STD_OUT" or "STD_ERR",
   *     Debug::Domain::output will be directed to the corresponding standard
   *     stream.
   *   - Otherwise, "xxx_output" will be considered as the path to a file on the
   *     filesystem, and the Debug::Domain::output will be directed to this
   *     file.
   */

  void enableFromPreferences();

  /**
   * Enables a particular Debug::Domain, specified by it Debug::Domain::ID.
   * @param domain The Debug::Domain::ID of the Debug::Domain to activate.
   */
  void enable(Domain::ID domain);
  /**
   * Disables a particular Debug::Domain, specified by it Debug::Domain::ID.
   * @param domain The Debug::Domain::ID of the Debug::Domain to disable.
   */
  void disable(Domain::ID domain);
  /**
   * \c true if the specified Debug::Domain is enabled
   * @param domain The Debug::Domain::ID of the Debug::Domain to test.
   * @return \c true if the Debug::Domain is enabled
   */
  bool enabled(Domain::ID domain);
  //@}
private:
  /**
   * \internal
   * The set of currently defined Debug::Domain.
   */
  typedef std::vector<Domain> SetOfDomain;
  /**
   * \internal
   * The set of currently opened file by the Debug singleton.
   */
  typedef std::set<std::ofstream*> SetOfFiles;
  /**
   * \internal
   * For convenience of intern programming. Here to not open a file
   * twice in enableFromPreferences() and enableFromEnvironment();
   */
  typedef std::map<std::string,std::ofstream*> MapFile;

  static Debug * instance_;


  SetOfDomain domains_;
  SetOfFiles files_;
  Debug();
  virtual ~Debug();

  /**
   * \internal
   * Adds Domain. will be copied by the SetOfDomain, so don't worry if its a
   * temporary variable. However there is some gcc compliance under certain
   * plateform if you do that, so its better for end-user to use the other
   * addDomain(bool, std::ostream &, const std::string &) method.
   * @param domain the Debug::Domain to add.
   * @return the unique Debug::Domain::ID associated with the Debug::Domain.
   */
  Domain::ID addDomain(Domain & domain);

  /**
   * \internal
   * Just perform the Job for a certain Debug::Domain when reading Preferences.
   */
  void setADomainAccordingPreferences(Domain&,MapFile & usedFile);

  /**
   * \internal
   * Gets a Debug::Domain .
   * @param id
   * @return
   */
  Domain & domain(Domain::ID id);

  class StreamBuf :public std::streambuf{

  };


};

#endif // LIBAMARSI_QUAD_DEBUG_H_
