/*
 * Copyright (c) 2010 Ecole Polytechnique Fédérale de Lausanne
 *
 * This file is part of libamarsi-quad.
 *
 * libamarsi-quad is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * libamarsi-quad is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with libamarsi-quad.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#ifndef LIBAMARSI_QUAD_PREFERENCE_READER_H_
#define LIBAMARSI_QUAD_PREFERENCE_READER_H_
#include <iostream>
#include <fstream>
#include <map>
#include <string>

#include <ostream>
#include <sstream>

/**
 * \defgroup Preferences Preferences Module.
 * Module that add Preferences in your code. Used extensively through
 * libamarsi-quad.
 *
 * \section use Use in libamarsi-quad
 * The preference module isused in amarsi quad. actually it reads the file
 * \c prefs/global.pref for some specific values. Some are intended to be
 * changed bythe user. But almost are values that affect the numerical stability
 *  of the model, and therefor should be left unchanged.
 *
 * We present a description of these variables :
 *
 * \subsection debug  Debug domain
 * The debug domain is used by the \ref Debug module to display output.Please
 * refer to the proper documentation. Only Debug::Domain::BASE and
 * Debug::Domain::CONTROLLER are currently used by the internal of
 * libamarsi-quad.
 *
 * \subsection cont Controller domain
 * In this section, only the \c initialialization_time is used. Its specifies
 * the time the model is left inactuated at the beginning of the simulation
 * and after a call to amarsi::Supervisor::resetSimulation(). Value below 2.0
 * are not recommanded
 *
 * \subsection knee  KneeServo domain
 *
 * \warning This section contains values that can dramatically changes the
 * behavior of the model/robot. Let them unchanged if you don't know what you
 * are doing.
 *
 *  - \c stiffness : stiffness of the compliant joint in N.rad^{-1}
 *  - \c damping : damping of the compliant joint in N.s.rad^{-1}
 *  - \c max_speed : maximal speed of the actuated joint.
 *  - \c max_torque : maximal torque of both servo-motor
 *  - \c max_compliant_velocity : maximal output speed of the servomotor of the
 *       compliant joint.
 *
 * \subsection toe  Toe domain
 * \warning Same as above, Do not change them unless you know what you are
 *          doing.
 *
 *  These value are here to define the hardness/softness of the contact joint
 *  made between the toe and the ground. these are directly the ODE ERP/CFM
 *  parameter given to the dContact structure .
 *
 * \subsection touch  TouchSensor
 *
 * This domain define the low-pass filter and the threshold that are used
 * on the norm of the value of the touchsensor to determine if there is contact
 * or not. The low-pass as the following equation, given the parameter \f$ a \f$
 * , input \f$ (x_n) \f$ and output \f$ (y_n) \f$:
 * \f[ y_i= \left (1 -a \right )* x_i + a * y_{i-1} \f]
 *
 *   - \c low_pass_amount : the parameter \f$ a \f$ in the equation above
 *   - \c threshold : the threshold used to determine if there is contact or not.
 *
 *
 * \ingroup user
 */
//@{
/**
 * This singleton aims to store configuration information for the
 * program.  The information are stored in .pref files, which are
 * simple text file. All the preferences variables are organized in
 * sections or PreferenceReader::Domain. Each variable has a unique
 * PreferenceReader::Domain.
 *
 * \section syntax Syntax of a .pref file
 * The syntax of a .pref file is really simple. Its defines Domain
 * with "[DomainName]" keyword at the beginning of a line. After such
 * definition every defined variable will be assigned to this Domain
 * until a new Domain definition is met. Domain could also be defined
 * in several part through the preference file.
 *
 * A preference variable is simply with "variable_name =
 * variable_value ", where variable_value is any string. Comment could
 * also be added with the '#' character. variable could be redefined
 * several time through the file, only the last seen value will be
 * remembered.
 *
 *
 * Here is an example of such a preference_file :
\code
[Domain1] #create a new domain
option1 = true
option2 = false
string1 = My super string #this comment will not be stored

# another comment here
[Domain2]
some_value = 1.42

[Domain1]
string1 =  My even-more super string #this value redefine previous value of Domain1.string1
\endcode
 *
 *
 * \section use Use of a Preference file
 *
 * To use the Preference system, you need to read one or several
 * preference file with the read() method. If you read several files,
 * definition will be appended in the read order, on the principle
 * "latest definition takes precedence".
 *
 * Then you could have access to the preference variables using the
 * getValue<T>(const std::string & domainName, const std::string & variableName, T & returnValue)
 * methods. Please notice that it is at this time that the value of
 * the variable is parsed to match the wanted type. parse for
 * std::string and numbered value function as usual. For bool values,
 * they are set to false if the string value match "false"
 * "disabled" or "0". Any other value will return \c true.
 *
 * When you explore preferences, you could choose between selecting a
 * Domain or not, by either calling
 * getValue<T>(const std::string & domainName, const std::string & variableName, T & returnValue) or
 * getValue<T>(const std::string & domainName, T & returnValue) const. If you
 * don't select one domain, then the last succesfully selected Domain
 * will be taken. Note that the selection of Domain could take time,
 * so if you have several value to extract in a Domain in a row, it
 * more efficient to select one Domain using setDomain(), and only
 * call getValue<T>(const std::string & domainName, T & returnValue) const access methods.
 *
 * \todo should throw exception, if needed.
 * \todo add type safe structure, that check malformed preference
 *       file, and faster access to the data.
 * \todo Should be renamed in "Preference", and also have output
 *        hability.
 */

class PreferenceReader{
  friend std::ostream &operator<<(std::ostream &out , const PreferenceReader &r);
  /**
   * \internal
   * Definition of a Domain
   */
  typedef std::map<std::string,std::string> Domain;
  /**
   * \internal
   * A list of Domain indexed by their name.
   */
  typedef std::map<std::string,Domain> ListOfDomain;

 public :

  static PreferenceReader& instance();

  //PreferenceReader(const std::string & file);
  virtual ~PreferenceReader();

  /**
   * Reads a new preference files. If one variable is redefined in the
   * new one, the new value will override the previous value.
   * @param file path to .pref file to read.
   */
  void read(const std::string & file);
  void read(std::istream & stream);
  /**
   * Removes any preferences definition.
   */
  void clear();

  /**
   * Returns the number of parsed files. \note this has been introduce
   * for a bad mechanism, consider it as obsolete.
   *
   * @return the number of parsed files.
   */
  unsigned int nbParsedFiles() const;

  /**
   * @name Access methods
   *
   * These methods are templated, therefore, when parsing the
   * preference file, string value are not parsed, but only
   * stored. This is only once one of these method is called then the
   * value is parsed. For standard type it is parsed using the
   * std::istringstream extractor. Note that for bool value, they are
   * setted to false if their value are either empty, "false",
   * "disabled" or "0", all other value return \c true. If you want to
   * extract more complex object, you can do it by providing the
   * correct extractor operator : std::istream operator<<(std::istream
   * &, MyObject&).
   */
  //@{
  /**
   * Sets the current Domain. while reading, all calls at
   * getValue<T>(const std::string & variableName, T& returnValue) const and
   * getValue<T>(const std::string & variableName) const will search values in the current
   * Domain. Any call to getValue<T>(const std::string& domainName, const std::string& variableName, T& returnValue)
   * or getValue<T>(const std::string & domainName, const std::string & variableName) may change the
   * current domain.
   *
   * @param name the name of the Domain to select.
   * @return \c true if the Domain exists (and therefore is selected).
   */
  bool setDomain(const std::string & name);
  /**
   * Gets a new value from the current Domain.
   *
   * @param name the variable name.
   * @param value a reference in which the parsed value will be
   *              stored.
   * @return \c true if the value exist in the Domain and had been
   *          correctly parsed.
   */
  template <typename T> bool getValue(const std::string & name,
                                      T & value) const;
  /**
   * Selects a new Domain and parse a given variable. \note If it
   * returns \c true, then the Domain has been changed, otherwise we
   * don't know. So please remind that after the call of this method,
   * the selected Domain may be uncertain.
   *
   * \todo should throw exception to leave the uncertainty when Domain
   * or variable is not found
   *
   * @param domainName the name of the Domain to select.
   * @param variableName the name of the variable to parse.
   * @param value the object to fill with the parsed value.
   * @return \c true if Domain succesfully selected and variable
   *         succesfully extracted.
   */
  template <typename T> bool getValue(const std::string & domainName,
                                      const std::string & variableName,
                                      T & value);
  /**
   * Gets a new variable from the current Domain. \note If the variable doesn't
   * exist, it will returns an object created with default Constructor
   * (one without any arguments)
   *
   * \todo should throw exception if the variable is not found.
   *
   * @param name the name of the variable to parse
   * @return the parsed value.
   */
  template <typename T> T getValue(const std::string & name) const;
  /**
   * Selects a new Domain and parse a given variable. \note If the
   * Domain doesn't exist or if the variable is not found, the
   * returned object will be one created with the default - without
   * any arguments - constructor.
   *
   * \todo should *really* throws exception when either the Domain is
   * not found or the variable doesn't exists
   *
   * @param domainName the name of the Domain to explore
   * @param name the name
   * @return
   */
  template <typename T> T getValue(const std::string & domainName,
                                   const std::string & name);
  //@}


  const Domain & currentDomainValues() const;
 private:
  //Disable the copyand construction
  PreferenceReader();
  PreferenceReader(const PreferenceReader & );
  PreferenceReader& operator=(PreferenceReader &);

  void init();
  void parse();


  std::istream *stream_;
  static PreferenceReader r;
  ListOfDomain domains_;
  ListOfDomain::const_iterator currentDomain_;
  unsigned int nbParsedFile_;

  static Domain s_voidDomain;

};

template <typename T>
  bool PreferenceReader::getValue(const std::string & name,
                                  T & value) const{
  Domain::const_iterator it= currentDomain_->second.find(name);
  if(it==currentDomain_->second.end())
    return false;
  std::istringstream iss(it->second);
  iss>>value;
  return true;
}

template <typename T>
  bool PreferenceReader::getValue(const std::string & domainName,
                                  const std::string & prefName,
                                  T & value){
  if(!this->setDomain(domainName))
    return false;
  return this->getValue(prefName,value);
}

template <typename T>
  T PreferenceReader::getValue(const std::string & name) const{
  	T tmp;
  	getValue(name,tmp);
  	return tmp;
  }

template <typename T>
	T PreferenceReader::getValue(const std::string & domainName,
	                             const std::string & name){
		T tmp;
  	getValue(domainName,name,tmp);
  	return tmp;
}

template <>
  inline bool PreferenceReader::getValue<bool>(const std::string & name ,
										bool & value) const {
  Domain::const_iterator it= currentDomain_->second.find(name);
  if(it==currentDomain_->second.end())
    return false;
  value=true;
  if(it->second.empty() || it->second=="false" ||
     it->second=="no"   || it->second=="disabled"){
    value=false;
    return true;
  }

  return true;
}
template <>
  inline bool PreferenceReader::getValue<std::string>(const std::string & name,
											   std::string & value) const{
  Domain::const_iterator it= currentDomain_->second.find(name);
  if(it==currentDomain_->second.end())
    return false;
  value=it->second;
  return true;
}
//@}
#endif // AMARSI_PREFERENCE_READER_H_
